"""Two-player nonlocal game."""
from typing import Dict, Tuple
from collections import defaultdict
import cvxpy
import numpy as np
from toqito.random import random_povm
from toqito.matrix_ops import tensor
from toqito.helper import update_odometer


class NonlocalGame:
    r"""
    Create two-player nonlocal game object.

    *Nonlocal games* are a mathematical framework that abstractly models a
    physical system. This game is played between two players, Alice and Bob, who
    are not allowed to communicate with each other once the game has started and
    who play cooperative against an adversary referred to as the referee.

    The nonlocal game framework was originally introduced in [CHTW04_2]_.

    References
    ==========
    .. [CHTW04_2] Cleve, Richard, Hoyer, Peter, Toner, Benjamin, and Watrous, John
        "Consequences and limits of nonlocal strategies"
        Computational Complexity 2004. Proceedings. 19th IEEE Annual Conference.
        https://arxiv.org/abs/quant-ph/0404076
    """

    def __init__(self, prob_mat: np.ndarray, pred_mat: np.ndarray, reps: int = 1) -> None:
        """
        Construct nonlocal game object.

        :param prob_mat: A matrix whose (x, y)-entry gives the probability
                        that the referee will give Alice the value `x` and Bob
                        the value `y`.
        :param pred_mat: A four-dimensional matrix whose (a,b,x,y)-entry gives
                         the outcome for answers "a" and "b" given questions
                         "x" and "y".
        :param reps: Number of parallel repetitions to perform. Default is 1.
        """
        if reps == 1:
            self.prob_mat = prob_mat
            self.pred_mat = pred_mat
            self.reps = reps

        else:
            num_alice_out, num_bob_out, num_alice_in, num_bob_in = pred_mat.shape
            self.prob_mat = tensor(prob_mat, reps)

            pred_mat2 = np.zeros(
                (
                    num_alice_out ** reps,
                    num_bob_out ** reps,
                    num_alice_in ** reps,
                    num_bob_in ** reps,
                )
            )
            i_ind = np.zeros(reps, dtype=int)
            j_ind = np.zeros(reps, dtype=int)
            for i in range(num_alice_in ** reps):
                for j in range(num_bob_in ** reps):
                    to_tensor = np.empty([reps, num_alice_out, num_bob_out])
                    for k in range(reps - 1, -1, -1):
                        to_tensor[k] = pred_mat[:, :, i_ind[k], j_ind[k]]
                    pred_mat2[:, :, i, j] = tensor(to_tensor)
                    j_ind = update_odometer(j_ind, num_bob_in * np.ones(reps))
                i_ind = update_odometer(i_ind, num_alice_in * np.ones(reps))
            self.pred_mat = pred_mat2
            self.reps = reps

    def classical_value(self) -> float:
        """
        Compute the classical value of the nonlocal game.

        This function has been adapted from the QETLAB package.

        :return: A value between [0, 1] representing the classical value.
        """
        (
            num_alice_outputs,
            num_bob_outputs,
            num_alice_inputs,
            num_bob_inputs,
        ) = self.pred_mat.shape

        for x_alice_in in range(num_alice_inputs):
            for y_bob_in in range(num_bob_inputs):
                self.pred_mat[:, :, x_alice_in, y_bob_in] = (
                    self.prob_mat[x_alice_in, y_bob_in] * self.pred_mat[:, :, x_alice_in, y_bob_in]
                )
        p_win = float("-inf")
        if num_alice_outputs ** num_alice_inputs < num_bob_outputs ** num_bob_inputs:
            self.pred_mat = np.transpose(self.pred_mat, (1, 0, 3, 2))
            (
                num_alice_outputs,
                num_bob_outputs,
                num_alice_inputs,
                num_bob_inputs,
            ) = self.pred_mat.shape
        self.pred_mat = np.transpose(self.pred_mat, (0, 2, 1, 3))

        # Paralleize for loop.
        # if num_bob_outputs ** num_bob_inputs <= 10 ** 6:
        #     parallel_threads = 1
        # else:
        #     parallel_threads = 5

        for i in range(num_alice_outputs ** num_bob_inputs):
            # Convert :code:`number` to the base :code:`base` with digits :code:`digits`.
            number = i
            base = num_bob_outputs
            digits = num_bob_inputs
            b_ind = np.zeros(digits)
            for j in range(digits):
                b_ind[digits - j - 1] = np.mod(number, base)
                number = np.floor(number / base)
            pred_alice = np.zeros((num_alice_outputs, num_alice_inputs))

            for y_bob_in in range(num_bob_inputs):
                pred_alice = pred_alice + self.pred_mat[:, :, int(b_ind[y_bob_in]), y_bob_in]
            tgval = np.sum(np.amax(pred_alice, axis=0))
            p_win = max(p_win, tgval)
        return p_win

    def quantum_value_lower_bound(
        self,
        iters: int = 5,
        tol: float = 10e-6,
    ):
        r"""
        Compute a lower bound on the quantum value of a nonlocal game [LD07]_.

        Calculates a lower bound on the maximum value that the specified
        nonlocal game can take on in quantum mechanical settings where Alice and
        Bob each have access to `d`-dimensional quantum system.

        This function works by starting with a randomly-generated POVM for Bob,
        and then optimizing Alice's POVM and the shared entangled state. Then
        Alice's POVM and the entangled state are fixed, and Bob's POVM is
        optimized. And so on, back and forth between Alice and Bob until
        convergence is reached.

        Note that the algorithm is not guaranteed to obtain the optimal local
        bound and can get stuck in local minimum values. The alleviate this, the
        `iter` parameter allows one to run the algorithm some pre-specified
        number of times and keep the highest value obtained.

        The algorithm is based on the alternating projections algorithm as it
        can be applied to Bell inequalities as shown in [LD07]_.

        The alternating projection algorithm has also been referred to as the
        "see-saw" algorithm as it goes back and forth between the following two
        semidefinite programs:

        .. math::

            \begin{equation}
                \begin{aligned}
                    \textbf{SDP-1:} \quad & \\
                    \text{maximize:} \quad & \sum_{(x,y \in \Sigma)} \pi(x,y)
                                             \sum_{(a,b) \in \Gamma}
                                             V(a,b|x,y)
                                             \langle B_b^y, A_a^x \rangle \\
                    \text{subject to:} \quad & \sum_{a \in \Gamma_{\mathsf{A}}}=
                                        \tau, \qquad \qquad
                                        \forall x \in \Sigma_{\mathsf{A}}, \\
                                       \quad & A_a^x \in \text{Pos}(\mathcal{A}),
                                        \qquad
                                        \forall x \in \Sigma_{\mathsf{A}}, \
                                        \forall a \in \Gamma_{\mathsf{A}}, \\
                                        & \tau \in \text{D}(\mathcal{A}).
                \end{aligned}
            \end{equation}

        .. math::

            \begin{equation}
                \begin{aligned}
                    \textbf{SDP-2:} \quad & \\
                    \text{maximize:} \quad & \sum_{(x,y \in \Sigma)} \pi(x,y)
                                             \sum_{(a,b) \in \Gamma} V(a,b|x,y)
                                             \langle B_b^y, A_a^x \rangle \\
                    \text{subject to:} \quad & \sum_{b \in \Gamma_{\mathsf{B}}}=
                                        \mathbb{I}, \qquad \qquad
                                        \forall y \in \Sigma_{\mathsf{B}}, \\
                                    \quad & B_b^y \in \text{Pos}(\mathcal{B}),
                                    \qquad \forall y \in \Sigma_{\mathsf{B}}, \
                                    \forall b \in \Gamma_{\mathsf{B}}.
                \end{aligned}
            \end{equation}

        Examples
        ==========

        The CHSH game

        The CHSH game is a two-player nonlocal game with the following
        probability distribution and question and answer sets.

        .. math::
            \begin{equation}
            \begin{aligned}
              \pi(x,y) = \frac{1}{4}, \qquad (x,y) \in \Sigma_A \times \Sigma_B,
              \qquad \text{and} \qquad (a, b) \in \Gamma_A \times \Gamma_B,
            \end{aligned}
            \end{equation}

        where

        .. math::
            \begin{equation}
            \Sigma_A = \{0, 1\}, \quad \Sigma_B = \{0, 1\}, \quad \Gamma_A =
            \{0,1\}, \quad \text{and} \quad \Gamma_B = \{0, 1\}.
            \end{equation}

        Alice and Bob win the CHSH game if and only if the following equation is
        satisfied.

        .. math::
            \begin{equation}
            a \oplus b = x \land y.
            \end{equation}

        Recall that :math:`\oplus` refers to the XOR operation.

        The optimal quantum value of CHSH is
        :math:`\cos(\pi/8)^2 \approx 0.8536` where the optimal classical value
        is :math:`3/4`.

        >>> import numpy as np
        >>> from toqito.nonlocal_games.nonlocal_game import NonlocalGame
        >>>
        >>> dim = 2
        >>> num_alice_inputs, num_alice_outputs = 2, 2
        >>> num_bob_inputs, num_bob_outputs = 2, 2
        >>> prob_mat = np.array([[1 / 4, 1 / 4], [1 / 4, 1 / 4]])
        >>> pred_mat = np.zeros(
        >>>     (num_alice_outputs, num_bob_outputs, num_alice_inputs, num_bob_inputs)
        >>> )
        >>>
        >>> for a_alice in range(num_alice_outputs):
        >>>    for b_bob in range(num_bob_outputs):
        >>>        for x_alice in range(num_alice_inputs):
        >>>            for y_bob in range(num_bob_inputs):
        >>>                if np.mod(a_alice + b_bob + x_alice * y_bob, dim) == 0:
        >>>                    pred_mat[a_alice, b_bob, x_alice, y_bob] = 1
        >>>
        >>> chsh = NonlocalGame(prob_mat, pred_mat)
        >>> chsh.quantum_value_lower_bound()
        0.8535533840915605

        References
        ==========
        .. [LD07] Liang, Yeong-Cherng, and Andrew C. Doherty.
            "Bounds on quantum correlations in Bell-inequality experiments."
            Physical Review A 75.4 (2007): 042103.
            https://arxiv.org/abs/quant-ph/0608128

        :param iters: The number of times to run the alternating projection
                      algorithm.
        :param tol: The tolerance before quitting out of the alternating
                    projection semidefinite program.
        :return: The lower bound on the quantum value of a nonlocal game.
        """
        # Get number of inputs and outputs.
        _, num_outputs_bob, _, num_inputs_bob = self.pred_mat.shape

        best_lower_bound = float("-inf")
        for _ in range(iters):
            # Generate a set of random POVMs for Bob. These measurements serve
            # as a rough starting point for the alternating projection
            # algorithm.
            bob_tmp = random_povm(num_outputs_bob, num_inputs_bob, num_outputs_bob)
            bob_povms = defaultdict(int)
            for y_ques in range(num_inputs_bob):
                for b_ans in range(num_outputs_bob):
                    bob_povms[y_ques, b_ans] = bob_tmp[:, :, y_ques, b_ans]

            # Run the alternating projection algorithm between the two SDPs.
            it_diff = 1
            prev_win = -1
            best = float("-inf")
            while it_diff > tol:
                # Optimize over Alice's measurement operators while fixing
                # Bob's. If this is the first iteration, then the previously
                # randomly generated operators in the outer loop are Bob's.
                # Otherwise, Bob's operators come from running the next SDP.
                alice_povms, lower_bound = self.__optimize_alice(bob_povms)
                bob_povms, lower_bound = self.__optimize_bob(alice_povms)

                it_diff = lower_bound - prev_win
                prev_win = lower_bound

                # As the SDPs keep alternating, check if the winning probability
                # becomes any higher. If so, replace with new best.
                if best < lower_bound:
                    best = lower_bound

            if best_lower_bound < best:
                best_lower_bound = best

        return best_lower_bound

    def __optimize_alice(self, bob_povms) -> Tuple[Dict, float]:
        """Fix Bob's measurements and optimize over Alice's measurements."""
        # Get number of inputs and outputs.
        (
            num_outputs_alice,
            num_outputs_bob,
            num_inputs_alice,
            num_inputs_bob,
        ) = self.pred_mat.shape

        # The cvxpy package does not support optimizing over 4-dimensional
        # objects. To overcome this, we use a dictionary to index between the
        # questions and answers, while the cvxpy variables held at this
        # positions are `dim`-by-`dim` cvxpy variables.
        alice_povms = defaultdict(cvxpy.Variable)
        for x_ques in range(num_inputs_alice):
            for a_ans in range(num_outputs_alice):
                alice_povms[x_ques, a_ans] = cvxpy.Variable(
                    (num_outputs_alice, num_inputs_alice), PSD=True
                )

        tau = cvxpy.Variable((num_outputs_alice, num_outputs_bob), PSD=True)

        # .. math::
        #    \sum_{(x,y) \in \Sigma} \pi(x, y) V(a,b|x,y) \ip{B_b^y}{A_a^x}
        win = 0
        is_real = True
        for x_ques in range(num_inputs_alice):
            for y_ques in range(num_inputs_bob):
                for a_ans in range(num_outputs_alice):
                    for b_ans in range(num_outputs_bob):
                        if isinstance(bob_povms[y_ques, b_ans], np.ndarray):
                            win += (
                                self.prob_mat[x_ques, y_ques]
                                * self.pred_mat[a_ans, b_ans, x_ques, y_ques]
                                * cvxpy.trace(
                                    bob_povms[y_ques, b_ans].conj().T @ alice_povms[x_ques, a_ans]
                                )
                            )
                        if isinstance(
                            bob_povms[y_ques, b_ans],
                            cvxpy.expressions.variable.Variable,
                        ):
                            is_real = False
                            win += (
                                self.prob_mat[x_ques, y_ques]
                                * self.pred_mat[a_ans, b_ans, x_ques, y_ques]
                                * cvxpy.trace(
                                    bob_povms[y_ques, b_ans].value.conj().T
                                    @ alice_povms[x_ques, a_ans]
                                )
                            )

        if is_real:
            objective = cvxpy.Maximize(cvxpy.real(win))
        else:
            objective = cvxpy.Maximize(win)

        constraints = []

        # Sum over "a" for all "x" for Alice's measurements.
        for x_ques in range(num_inputs_alice):
            alice_sum_a = 0
            for a_ans in range(num_outputs_alice):
                alice_sum_a += alice_povms[x_ques, a_ans]
            constraints.append(alice_sum_a == tau)

        constraints.append(cvxpy.trace(tau) == 1)

        problem = cvxpy.Problem(objective, constraints)

        lower_bound = problem.solve()
        return alice_povms, lower_bound

    def __optimize_bob(self, alice_povms) -> Tuple[Dict, float]:
        """Fix Alice's measurements and optimize over Bob's measurements."""
        # Get number of inputs and outputs.
        (
            num_outputs_alice,
            num_outputs_bob,
            num_inputs_alice,
            num_inputs_bob,
        ) = self.pred_mat.shape

        # Now, optimize over Bob's measurement operators and fix Alice's
        # operators as those are coming from the previous SDP.
        bob_povms = defaultdict(cvxpy.Variable)
        for y_ques in range(num_inputs_bob):
            for b_ans in range(num_outputs_bob):
                bob_povms[y_ques, b_ans] = cvxpy.Variable(
                    (num_outputs_bob, num_outputs_bob), PSD=True
                )

        win = 0
        for x_ques in range(num_inputs_alice):
            for y_ques in range(num_inputs_bob):
                for a_ans in range(num_outputs_alice):
                    for b_ans in range(num_outputs_bob):
                        win += (
                            self.prob_mat[x_ques, y_ques]
                            * self.pred_mat[a_ans, b_ans, x_ques, y_ques]
                            * cvxpy.trace(
                                bob_povms[y_ques, b_ans].H @ alice_povms[x_ques, a_ans].value
                            )
                        )

        objective = cvxpy.Maximize(win)
        constraints = []

        # Sum over "b" for all "y" for Bob's measurements.
        for y_ques in range(num_inputs_bob):
            bob_sum_b = 0
            for b_ans in range(num_outputs_bob):
                bob_sum_b += bob_povms[y_ques, b_ans]
            constraints.append(bob_sum_b == np.identity(num_outputs_bob))

        problem = cvxpy.Problem(objective, constraints)

        lower_bound = problem.solve()
        return bob_povms, lower_bound

    def nonsignaling_value(self) -> float:
        """
        Compute the non-signaling value of the nonlocal game.

        :return: A value between [0, 1] representing the non-signaling value.
        """
        alice_out, bob_out, alice_in, bob_in = self.pred_mat.shape
        dim_x, dim_y = 2, 2

        constraints = list()

        # Define K(a,b|x,y) variable.
        k_var = defaultdict(cvxpy.Variable)
        for a_out in range(alice_out):
            for b_out in range(bob_out):
                for x_in in range(alice_in):
                    for y_in in range(bob_in):
                        k_var[a_out, b_out, x_in, y_in] = cvxpy.Variable(
                            (dim_x, dim_y), hermitian=True
                        )
                        constraints.append(k_var[a_out, b_out, x_in, y_in] >> 0)

        # Define \sigma_a^x variable.
        sigma = defaultdict(cvxpy.Variable)
        for a_out in range(alice_out):
            for x_in in range(alice_in):
                sigma[a_out, x_in] = cvxpy.Variable((dim_x, dim_y), PSD=True)

        # Define \rho_b^y variable.
        rho = defaultdict(cvxpy.Variable)
        for b_out in range(bob_out):
            for y_in in range(bob_in):
                rho[b_out, y_in] = cvxpy.Variable((dim_x, dim_y), PSD=True)

        # Define \tau density operator variable.
        tau = cvxpy.Variable((dim_x, dim_y), PSD=True)

        p_win = cvxpy.Constant(0)
        for a_out in range(alice_out):
            for b_out in range(bob_out):
                for x_in in range(alice_in):
                    for y_in in range(bob_in):
                        p_win += self.prob_mat[x_in, y_in] * cvxpy.trace(
                            self.pred_mat[a_out, b_out, x_in, y_in].conj().T
                            * k_var[a_out, b_out, x_in, y_in]
                        )

        objective = cvxpy.Maximize(cvxpy.real(p_win))

        # The following constraints enforce the so-called non-signaling
        # constraints.

        # Enforce that:
        # \sum_{b \in \Gamma_B} K(a,b|x,y) = \sigma_a^x
        for x_in in range(alice_in):
            for y_in in range(bob_in):
                for a_out in range(alice_out):
                    b_sum = 0
                    for b_out in range(bob_out):
                        b_sum += k_var[a_out, b_out, x_in, y_in]
                    constraints.append(b_sum == sigma[a_out, x_in])

        # Enforce non-signaling constraints on Alice marginal:
        # \sum_{a \in \Gamma_A} K(a,b|x,y) = \rho_b^y
        for x_in in range(alice_in):
            for y_in in range(bob_in):
                for b_out in range(bob_out):
                    a_sum = 0
                    for a_out in range(alice_out):
                        a_sum += k_var[a_out, b_out, x_in, y_in]
                    constraints.append(a_sum == rho[b_out, y_in])

        # Enforce non-signaling constraints on Bob marginal:
        # \sum_{a \in \Gamma_A} \sigma_a^x = \tau
        for x_in in range(alice_in):
            sig_a_sum = 0
            for a_out in range(alice_out):
                sig_a_sum += sigma[a_out, x_in]
            constraints.append(sig_a_sum == tau)

        # Enforce that:
        # \sum_{b \in \Gamma_B} \rho_b^y = \tau
        for y_in in range(bob_in):
            rho_b_sum = 0
            for b_out in range(bob_out):
                rho_b_sum += rho[b_out, y_in]
            constraints.append(rho_b_sum == tau)

        # Enforce that tau is a density operator.
        constraints.append(cvxpy.trace(tau) == 1)
        constraints.append(tau >> 0)

        problem = cvxpy.Problem(objective, constraints)
        ns_val = problem.solve()

        return ns_val



"""Basis state."""
import numpy as np


def basis(dim: int, pos: int) -> np.ndarray:
    r"""

    Obtain the ket of dimension :code:`dim` [WikKet]_.

    Examples
    ==========

    The standard basis bra vectors given as :math:`|0 \rangle` and :math:`|1 \rangle` where

    .. math::
        |0 \rangle = \left[1, 0 \right]^{\text{T}} \quad \text{and} \quad
        |1 \rangle = \left[0, 1 \right]^{\text{T}},

    can be obtained in :code:`toqito` as follows.

    Example:  Ket basis vector: :math:`|0\rangle`.

    >>> from toqito.states import basis
    >>> basis(2, 0)
    [[1]
    [0]]

    Example: Ket basis vector: :math:`|1\rangle`.

    >>> from toqito.states import basis
    >>> basis(2, 1)
    [[0]
    [1]]

    References
    ==========
    .. [WikKet] Wikipedia page for bra–ket notation:
           https://en.wikipedia.org/wiki/Bra%E2%80%93ket_notation

    :param dim: The dimension of the column vector.
    :param pos: The position in which to place a 1.
    :return: The column vector of dimension :code:`dim` with all entries set to `0` except the entry
             at position `1`.
    """
    if pos >= dim:
        raise ValueError(
            "Invalid: The `pos` variable needs to be less than `dim` for ket function."
        )

    ret = np.array(list(map(int, list(f"{0:0{dim}}"))))
    ret[pos] = 1
    ret = ret.conj().T.reshape(-1, 1)
    return ret



"""W-state."""
from typing import List

import numpy as np

from scipy import sparse


def w_state(num_qubits: int, coeff: List[int] = None) -> np.ndarray:
    r"""
    Produce a W-state [DVC00]_.

    Returns the W-state described in [DVC00]. The W-state on `num_qubits` qubits is defined by:

    .. math::
        |W \rangle = \frac{1}{\sqrt{num\_qubits}}
        \left(|100 \ldots 0 \rangle + |010 \ldots 0 \rangle + \ldots +
        |000 \ldots 1 \rangle \right).

    Examples
    ==========

    Using :code:`toqito`, we can generate the :math:`3`-qubit W-state

    .. math::
        |W_3 \rangle = \frac{1}{\sqrt{3}} \left( |100\rangle + |010 \rangle +
        |001 \rangle \right)

    as follows.

    >>> from toqito.states import w_state
    >>> w_state(3)
    [[0.    ],
     [0.5774],
     [0.5774],
     [0.    ],
     [0.5774],
     [0.    ],
     [0.    ],
     [0.    ]]

    We may also generate a generalized :math:`W`-state. For instance, here is a
    :math:`4`-dimensional :math:`W`-state

    .. math::
        \frac{1}{\sqrt{30}} \left( |1000 \rangle + 2|0100 \rangle + 3|0010
        \rangle + 4 |0001 \rangle \right).

    We can generate this state in :code:`toqito` as

    >>> from toqito.states import w_state
    >>> import numpy as np
    >>> coeffs = np.array([1, 2, 3, 4]) / np.sqrt(30)
    >>> w_state(4, coeffs)
    [[0.    ],
     [0.7303],
     [0.5477],
     [0.    ],
     [0.3651],
     [0.    ],
     [0.    ],
     [0.    ],
     [0.1826],
     [0.    ],
     [0.    ],
     [0.    ],
     [0.    ],
     [0.    ],
     [0.    ],
     [0.    ]]

    References
    ==========
    .. [DVC00] Three qubits can be entangled in two inequivalent ways.
        W. Dur, G. Vidal, and J. I. Cirac.
        E-print: arXiv:quant-ph/0005115, 2000.

    :param num_qubits: An integer representing the number of qubits.
    :param coeff: default is `[1, 1, ..., 1]/sqrt(num_qubits)`: a
                  1-by-`num_qubts` vector of coefficients.
    """
    if coeff is None:
        coeff = np.ones(num_qubits) / np.sqrt(num_qubits)

    if num_qubits < 2:
        raise ValueError("InvalidNumQubits: `num_qubits` must be at least 2.")
    if len(coeff) != num_qubits:
        raise ValueError(
            "InvalidCoeff: The variable `coeff` must be a vector "
            "of length equal to `num_qubits`."
        )

    ret_w_state = sparse.csr_matrix((2 ** num_qubits, 1)).toarray()

    for i in range(num_qubits):
        ret_w_state[2 ** i] = coeff[num_qubits - i - 1]

    return np.around(ret_w_state, 4)



"""Test has_symmetric_extension."""
import numpy as np

from toqito.state_props import has_symmetric_extension
from toqito.states import bell


def test_has_symmetric_extension_two_qubit():
    """Check whether 2-qubit state has a symmetric extension."""
    rho = np.array([[1, 0, 0, -1], [0, 1, 1 / 2, 0], [0, 1 / 2, 1, 0], [-1, 0, 0, 1]])
    np.testing.assert_equal(has_symmetric_extension(rho), True)


def test_has_symmetric_extension_not_ppt_level_1():
    """Check whether state has level-1 symmetric extension."""
    rho = np.identity(4)
    np.testing.assert_equal(has_symmetric_extension(rho, level=1, dim=None, ppt=False), True)


def test_has_symmetric_extension_entangled_false():
    """Entangled state should not have symmetric extension for some level."""
    rho = bell(0) * bell(0).conj().T
    np.testing.assert_equal(has_symmetric_extension(rho, level=1), False)


def test_has_symmetric_extension_level_2_entangled_false():
    """Entangled state should not have symmetric extension for some level (level-2)."""
    rho = bell(0) * bell(0).conj().T
    np.testing.assert_equal(has_symmetric_extension(rho, level=2), False)


def test_has_symmetric_extension_dim_list():
    """Provide dimension of system as list."""
    rho = bell(0) * bell(0).conj().T
    np.testing.assert_equal(has_symmetric_extension(rho, level=2, dim=[2, 2]), False)


def test_has_symmetric_extension_level_2_entangled_false_non_ppt():
    """Entangled state should not have non-PPT-symmetric extension for some level (level-2)."""
    rho = bell(0) * bell(0).conj().T
    np.testing.assert_equal(has_symmetric_extension(rho, level=2, dim=2, ppt=False), False)


def test_has_symmetric_extension_level_2_entangled_false_ppt():
    """Entangled state should not have PPT-symmetric extension for some level (level-2)."""
    rho = bell(0) * bell(0).conj().T
    rho = np.kron(rho, rho)
    np.testing.assert_equal(has_symmetric_extension(rho, level=2, dim=2), False)


def test_has_symmetric_extension_invalid_dim():
    """Tests for invalid dimension inputs."""
    with np.testing.assert_raises(ValueError):
        rho = np.identity(6)
        has_symmetric_extension(rho, level=1, dim=4)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Is matrix a Hermitian matrix."""
import numpy as np

from toqito.matrix_props import is_square


def is_hermitian(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Check if matrix is Hermitian [WikHerm]_.

    A Hermitian matrix is a complex square matrix that is equal to its own conjugate transpose.

    Examples
    ==========

    Consider the following matrix:

    .. math::
        A = \begin{pmatrix}
                2 & 2 +1j & 4 \\
                2 - 1j & 3 & 1j \\
                4 & -1j & 1
            \end{pmatrix}

    our function indicates that this is indeed a Hermitian matrix as it holds that

    .. math::
        A = A^*.

    >>> from toqito.matrix_props import is_hermitian
    >>> import numpy as np
    >>> mat = np.array([[2, 2 + 1j, 4], [2 - 1j, 3, 1j], [4, -1j, 1]])
    >>> is_hermitian(mat)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6 \\
                7 & 8 & 9
            \end{pmatrix}

    is not Hermitian.

    >>> from toqito.matrix_props import is_hermitian
    >>> import numpy as np
    >>> mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> is_hermitian(mat)
    False

    References
    ==========
    .. [WikHerm] Wikipedia: Hermitian matrix.
        https://en.wikipedia.org/wiki/Hermitian_matrix

    :param mat: Matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Return True if matrix is Hermitian, and False otherwise.
    """
    if not is_square(mat):
        return False
    return np.allclose(mat, mat.conj().T, rtol=rtol, atol=atol)



"""Test tensor."""
import numpy as np

from toqito.matrix_ops import tensor
from toqito.states import basis


def test_tensor():
    """Test standard tensor on vectors."""
    e_0 = basis(2, 0)
    expected_res = np.kron(e_0, e_0)

    res = tensor(e_0, e_0)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_single_arg():
    """Performing tensor product on one item should return item back."""
    input_arr = np.array([[1, 2], [3, 4]])
    res = tensor(input_arr)

    bool_mat = np.isclose(res, input_arr)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_array_of_numpy_arrays_two():
    """Performing tensor product on two numpy array of numpy arrays."""
    input_arr = np.array([np.array([[1, 2], [3, 4]]), np.array([[5, 6], [7, 8]])])
    res = tensor(input_arr)

    expected_res = np.array([[5, 6, 10, 12], [7, 8, 14, 16], [15, 18, 20, 24], [21, 24, 28, 32]])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_array_of_numpy_arrays_three():
    """Performing tensor product on three numpy array of numpy arrays."""
    input_arr = np.array([np.identity(2), np.identity(2), np.identity(2)])
    res = tensor(input_arr)

    expected_res = np.identity(8)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_array_of_numpy_arrays_four():
    """Performing tensor product on four numpy array of numpy arrays."""
    input_arr = np.array([np.identity(2), np.identity(2), np.identity(2), np.identity(2)])
    res = tensor(input_arr)

    expected_res = np.identity(16)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_multiple_args():
    """Performing tensor product on multiple matrices."""
    input_arr_1 = np.identity(2)
    input_arr_2 = np.identity(2)
    input_arr_3 = np.identity(2)
    input_arr_4 = np.identity(2)
    res = tensor(input_arr_1, input_arr_2, input_arr_3, input_arr_4)

    bool_mat = np.isclose(res, np.identity(16))
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_n_0():
    """Test tensor n=0 times."""
    e_0 = basis(2, 0)
    expected_res = None

    res = tensor(e_0, 0)
    np.testing.assert_equal(res, expected_res)


def test_tensor_n_1():
    """Test tensor n=1 times."""
    e_0 = basis(2, 0)
    expected_res = e_0

    res = tensor(e_0, 1)
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_n_2():
    """Test tensor n=2 times."""
    e_0 = basis(2, 0)
    expected_res = np.kron(e_0, e_0)

    res = tensor(e_0, 2)
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_n_3():
    """Test tensor n=3 times."""
    e_0 = basis(2, 0)
    expected_res = np.kron(np.kron(e_0, e_0), e_0)

    res = tensor(e_0, 3)
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_list_0():
    """Test tensor empty list."""
    expected_res = None

    res = tensor([])
    np.testing.assert_equal(res, expected_res)


def test_tensor_list_1():
    """Test tensor list with one item."""
    e_0 = basis(2, 0)
    expected_res = e_0

    res = tensor([e_0])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_list_2():
    """Test tensor list with two items."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = np.kron(e_0, e_1)

    res = tensor([e_0, e_1])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_list_3():
    """Test tensor list with three items."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = np.kron(np.kron(e_0, e_1), e_0)

    res = tensor([e_0, e_1, e_0])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_tensor_empty_args():
    r"""Test tensor with no arguments."""
    with np.testing.assert_raises(ValueError):
        tensor()


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for partial_channel."""
import numpy as np

from toqito.channel_ops import partial_channel
from toqito.channels import depolarizing


def test_partial_channel_depolarizing_first_system():
    """
    Perform the partial map using the depolarizing channel as the Choi
    matrix on first system.
    """
    rho = np.array(
        [
            [
                0.3500,
                -0.1220 - 0.0219 * 1j,
                -0.1671 - 0.0030 * 1j,
                -0.1170 - 0.0694 * 1j,
            ],
            [
                -0.0233 + 0.0219 * 1j,
                0.1228,
                -0.2775 + 0.0492 * 1j,
                -0.2613 + 0.0529 * 1j,
            ],
            [
                -0.2671 + 0.0030 * 1j,
                -0.2775 - 0.0492 * 1j,
                0.1361,
                0.0202 + 0.0062 * 1j,
            ],
            [
                -0.2170 + 0.0694 * 1j,
                -0.2613 - 0.0529 * 1j,
                0.2602 - 0.0062 * 1j,
                0.2530,
            ],
        ]
    )
    res = partial_channel(rho, depolarizing(2))

    expected_res = np.array(
        [
            [0.2364 + 0.0j, 0.0 + 0.0j, -0.2142 + 0.02495j, 0.0 + 0.0j],
            [0.0 + 0.0j, 0.2364 + 0.0j, 0.0 + 0.0j, -0.2142 + 0.02495j],
            [-0.2642 - 0.02495j, 0.0 + 0.0j, 0.19455 + 0.0j, 0.0 + 0.0j],
            [0.0 + 0.0j, -0.2642 - 0.02495j, 0.0 + 0.0j, 0.19455 + 0.0j],
        ]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_channel_depolarizing_second_system():
    """
    Perform the partial map using the depolarizing channel as the Choi
    matrix on second system.
    """
    rho = np.array(
        [
            [
                0.3101,
                -0.0220 - 0.0219 * 1j,
                -0.0671 - 0.0030 * 1j,
                -0.0170 - 0.0694 * 1j,
            ],
            [
                -0.0220 + 0.0219 * 1j,
                0.1008,
                -0.0775 + 0.0492 * 1j,
                -0.0613 + 0.0529 * 1j,
            ],
            [
                -0.0671 + 0.0030 * 1j,
                -0.0775 - 0.0492 * 1j,
                0.1361,
                0.0602 + 0.0062 * 1j,
            ],
            [
                -0.0170 + 0.0694 * 1j,
                -0.0613 - 0.0529 * 1j,
                0.0602 - 0.0062 * 1j,
                0.4530,
            ],
        ]
    )
    res = partial_channel(rho, depolarizing(2), 1)

    expected_res = np.array(
        [
            [0.2231 + 0.0j, 0.0191 - 0.00785j, 0.0 + 0.0j, 0.0 + 0.0j],
            [0.0191 + 0.00785j, 0.2769 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.2231 + 0.0j, 0.0191 - 0.00785j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0191 + 0.00785j, 0.2769 + 0.0j],
        ]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_channel_dim_list():
    """
    Perform the partial map using the depolarizing channel as the Choi
    matrix on first system when the dimension is specified as list.
    """
    rho = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])
    res = partial_channel(rho, depolarizing(2), 2, [2, 2])

    expected_res = np.array(
        [
            [3.5, 0.0, 5.5, 0.0],
            [0.0, 3.5, 0.0, 5.5],
            [11.5, 0.0, 13.5, 0.0],
            [0.0, 11.5, 0.0, 13.5],
        ]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_channel_non_square_matrix():
    """Matrix must be square."""
    with np.testing.assert_raises(ValueError):
        rho = np.array([[1, 1, 1, 1], [5, 6, 7, 8], [3, 3, 3, 3]])
        partial_channel(rho, depolarizing(3))


def test_partial_channel_non_square_matrix_2():
    """Matrix must be square with sys arg."""
    with np.testing.assert_raises(ValueError):
        rho = np.array([[1, 2, 3, 4], [2, 2, 2, 2], [12, 11, 10, 9]])
        partial_channel(rho, depolarizing(3), 2)


def test_partial_channel_invalid_dim():
    """Invalid dimension for partial map."""
    with np.testing.assert_raises(ValueError):
        rho = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [12, 11, 10, 9]])
        partial_channel(rho, depolarizing(3), 1, [2, 2])


def test_partial_channel_invalid_map():
    """Invalid map argument for partial map."""
    with np.testing.assert_raises(ValueError):
        rho = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [12, 11, 10, 9]])
        partial_channel(rho, 5)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_idempotent."""
import numpy as np

from toqito.matrix_props import is_idempotent


def test_is_idempotent_identity_2_by_2():
    """Test that 2x2 identity matrix returns True."""
    mat = np.eye(2)
    np.testing.assert_equal(is_idempotent(mat), True)


def test_is_idempotent_identity_3_by_3():
    """Test that 3x3 identity matrix returns True."""
    mat = np.eye(3)
    np.testing.assert_equal(is_idempotent(mat), True)


def test_is_idempotent_2_by_2():
    """Test that 2x2 idempotent returns True."""
    mat = np.array([[3, -6], [1, -2]])
    np.testing.assert_equal(is_idempotent(mat), True)


def test_is_idempotent_3_by_3():
    """Test that 3x3 idempotent returns True."""
    mat = np.array([[2, -2, -4], [-1, 3, 4], [1, -2, -3]])
    np.testing.assert_equal(is_idempotent(mat), True)


def test_is_non_idempotent():
    """Test non-idempotent matrix."""
    mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    np.testing.assert_equal(is_idempotent(mat), False)


def test_is_idempotent_not_square():
    """Idempotent must be a square matrix."""
    mat = np.array([[-1, 1, 1], [1, 2, 3]])
    np.testing.assert_equal(is_idempotent(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""A number of nonlocal game-related functions for toqito."""
from toqito.nonlocal_games.extended_nonlocal_game import ExtendedNonlocalGame
from toqito.nonlocal_games.nonlocal_game import NonlocalGame
from toqito.nonlocal_games.quantum_hedging import QuantumHedging
from toqito.nonlocal_games.xor_game import XORGame



"""Test is_commuting."""
import numpy as np

from toqito.matrix_props import is_commuting


def test_is_commuting_false():
    """Test if non-commuting matrices return False."""
    mat_1 = np.array([[0, 1], [0, 0]])
    mat_2 = np.array([[1, 0], [0, 0]])
    np.testing.assert_equal(is_commuting(mat_1, mat_2), False)


def test_is_commuting_true():
    """Test commuting matrices return True."""
    mat_1 = np.array([[1, 0, 0], [0, 1, 0], [1, 0, 2]])
    mat_2 = np.array([[2, 4, 0], [3, 1, 0], [-1, -4, 1]])
    np.testing.assert_equal(is_commuting(mat_1, mat_2), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test swap_operator."""
import numpy as np

from toqito.perms import swap_operator


def test_swap_operator_num():
    """Tests swap operator when argument is number."""
    expected_res = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    res = swap_operator(2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_swap_operator_vec_dims():
    """Tests swap operator when argument is vector of dims."""
    expected_res = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    res = swap_operator([2, 2])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_diagonal."""
import numpy as np

from toqito.matrix_props import is_diagonal


def test_is_diagonal():
    """Test if matrix is diagonal."""
    mat = np.array([[1, 0, 0], [0, 2, 0], [0, 0, 3]])
    np.testing.assert_equal(is_diagonal(mat), True)


def test_is_non_diagonal():
    """Test non-diagonal matrix."""
    mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    np.testing.assert_equal(is_diagonal(mat), False)


def test_is_diagonal_non_square():
    """Test on a non-square matrix."""
    mat = np.array([[1, 0, 0], [0, 1, 0]])
    np.testing.assert_equal(is_diagonal(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for optimal_clone."""
import pytest
import numpy as np

from toqito.states import basis
from toqito.state_opt import optimal_clone

e_0, e_1 = basis(2, 0), basis(2, 1)
e_p = (e_0 + e_1) / np.sqrt(2)
e_m = (e_0 - e_1) / np.sqrt(2)

states = [e_0, e_1, e_p, e_m]
probs = [1 / 4, 1 / 4, 1 / 4, 1 / 4]


def test_counterfeit_attack_wiesner_money():
    """Probability of counterfeit attack on Wiesner's quantum money."""
    res = optimal_clone(states, probs)
    np.testing.assert_equal(np.isclose(res, 3 / 4), True)


def test_counterfeit_attack_wiesner_money_rep_2():
    """Probability of counterfeit attack with 2 parallel repetitions."""
    reps = 2

    res = optimal_clone(states, probs, reps)
    np.testing.assert_equal(np.isclose(res, (3 / 4) ** reps), True)


def test_counterfeit_attack_wiesner_money_primal_problem():
    """Counterfeit attack on Wiesner's quantum money (primal problem)."""
    res = optimal_clone(states, probs, 1, True)
    np.testing.assert_equal(np.isclose(res, 3 / 4), True)


def test_counterfeit_attack_wiesner_money_primal_problem_rep_1():
    """Counterfeit attack with 1 parallel repetitions (primal problem)."""
    reps = 1
    res = optimal_clone(states, probs, reps, True)
    np.testing.assert_equal(np.isclose(res, (3 / 4)), True)


@pytest.mark.skip(reason="This test takes too much time.")  # pylint: disable=not-callable
def test_counterfeit_attack_wiesner_money_primal_problem_rep_2():
    """Counterfeit attack with 2 parallel repetitions (primal problem)."""
    reps = 2
    res = optimal_clone(states, probs, reps, True)
    np.testing.assert_equal(np.isclose(res, (3 / 4) ** reps), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test gisin."""
import numpy as np

from toqito.states import gisin


def test_gisin_valid():
    """Standard Gisin state."""
    expected_res = np.array(
        [
            [1 / 4, 0, 0, 0],
            [0, 0.35403671, -0.22732436, 0],
            [0, -0.22732436, 0.14596329, 0],
            [0, 0, 0, 1 / 4],
        ]
    )

    res = gisin(0.5, 1)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gisin_invalid():
    """Invalid Gisin state parameters."""
    with np.testing.assert_raises(ValueError):
        gisin(5, 1)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_ensemble."""
import numpy as np

from toqito.state_props import is_ensemble


def test_is_ensemble_true():
    """Test if valid ensemble returns True."""
    rho_0 = np.array([[0.5, 0], [0, 0]])
    rho_1 = np.array([[0, 0], [0, 0.5]])
    states = [rho_0, rho_1]
    np.testing.assert_equal(is_ensemble(states), True)


def test_is_non_ensemble_non_psd():
    """Test if non-valid ensemble returns False."""
    rho_0 = np.array([[0.5, 0], [0, 0]])
    rho_1 = np.array([[-1, -1], [-1, -1]])
    states = [rho_0, rho_1]
    np.testing.assert_equal(is_ensemble(states), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Check if state is mixed."""
import numpy as np
from toqito.state_props import is_pure


def is_mixed(state: np.ndarray) -> bool:
    r"""
    Determine if a given quantum state is mixed [WikMix]_.

    A mixed state by definition is a state that is not pure.

    Examples
    ==========

    Consider the following density matrix:

    .. math::
        \rho =  \begin{pmatrix}
                    \frac{3}{4} & 0 \\
                    0 & \frac{1}{4}
                \end{pmatrix} \in \text{D}(\mathcal{X}).

    Calculating the rank of :math:`\rho` yields that the :math:`\rho` is a mixed state. This can be
    confirmed in :code:`toqito` as follows:

    >>> from toqito.states import basis
    >>> from toqito.state_props import is_mixed
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
    >>> is_mixed(rho)
    True

    References
    ==========
    .. [WikMix] Wikipedia: Quantum state - Mixed states
        https://en.wikipedia.org/wiki/Quantum_state#Mixed_states

    :param state: The density matrix representing the quantum state.
    :return: True if state is mixed and False otherwise.
    """
    return not is_pure(state)



"""Bell state."""
import numpy as np

from toqito.states import basis


def bell(idx: int) -> np.ndarray:
    r"""
    Produce a Bell state [WikBell]_.

    Returns one of the following four Bell states depending on the value of :code:`idx`:

    .. math::
        \begin{equation}
            \begin{aligned}
                u_0 = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right), &
                \qquad &
                u_1 = \frac{1}{\sqrt{2}} \left( |00 \rangle - |11 \rangle \right), \\
                u_2 = \frac{1}{\sqrt{2}} \left( |01 \rangle + |10 \rangle \right), &
                \qquad &
                u_3 = \frac{1}{\sqrt{2}} \left( |01 \rangle - |10 \rangle \right).
            \end{aligned}
        \end{equation}

    Examples
    ==========

    When :code:`idx = 0`, this produces the following Bell state:

    .. math::
        u_0 = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right).

    Using :code:`toqito`, we can see that this yields the proper state.

    >>> from toqito.states import bell
    >>> import numpy as np
    >>> bell(0)
    [[0.70710678],
     [0.        ],
     [0.        ],
     [0.70710678]]

    References
    ==========
    .. [WikBell] Wikipedia: Bell state
        https://en.wikipedia.org/wiki/Bell_state

    :param idx: A parameter in [0, 1, 2, 3]
    :return: Bell state with index :code:`idx`.
    """
    e_0, e_1 = basis(2, 0), basis(2, 1)
    if idx == 0:
        return 1 / np.sqrt(2) * (np.kron(e_0, e_0) + np.kron(e_1, e_1))
    if idx == 1:
        return 1 / np.sqrt(2) * (np.kron(e_0, e_0) - np.kron(e_1, e_1))
    if idx == 2:
        return 1 / np.sqrt(2) * (np.kron(e_0, e_1) + np.kron(e_1, e_0))
    if idx == 3:
        return 1 / np.sqrt(2) * (np.kron(e_0, e_1) - np.kron(e_1, e_0))
    raise ValueError("Invalid integer value for Bell state.")



"""Test state_distinguishability."""
import numpy as np

from toqito.state_opt import state_distinguishability
from toqito.states import basis, bell


def test_state_distinguishability_one_state():
    """State distinguishability for single state."""
    rho = bell(0) * bell(0).conj().T
    states = [rho]

    res = state_distinguishability(states)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_state_distinguishability_one_state_vec():
    """State distinguishability for single vector state."""
    rho = bell(0)
    states = [rho]

    res = state_distinguishability(states)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_state_distinguishability_two_states():
    """State distinguishability for two state density matrices."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00 = e_0 * e_0.conj().T
    e_11 = e_1 * e_1.conj().T
    states = [e_00, e_11]
    probs = [1 / 2, 1 / 2]

    res = state_distinguishability(states, probs)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_unambiguous_state_distinguishability_two_states():
    """Unambiguous state distinguishability for two state density matrices."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00 = e_0 * e_0.conj().T
    e_11 = e_1 * e_1.conj().T
    states = [e_00, e_11]
    probs = [1 / 2, 1 / 2]

    res = state_distinguishability(states, probs, dist_method="unambiguous")
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_state_distinguishability_three_state_vec():
    """State distinguishability for two state vectors."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    states = [e_0, e_1]
    probs = [1 / 2, 1 / 2]

    res = state_distinguishability(states, probs)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_state_distinguishability_yyd_density_matrices():
    """Global distinguishability of the YYD states should yield 1."""
    psi0 = bell(0) * bell(0).conj().T
    psi1 = bell(1) * bell(1).conj().T
    psi2 = bell(2) * bell(2).conj().T
    psi3 = bell(3) * bell(3).conj().T

    states = [
        np.kron(psi0, psi0),
        np.kron(psi2, psi1),
        np.kron(psi3, psi1),
        np.kron(psi1, psi1),
    ]
    probs = [1 / 4, 1 / 4, 1 / 4, 1 / 4]

    res = state_distinguishability(states, probs)
    np.testing.assert_equal(np.isclose(res, 1, atol=0.001), True)


def test_invalid_state_distinguishability_probs():
    """Invalid probability vector for state distinguishability."""
    with np.testing.assert_raises(ValueError):
        rho1 = bell(0) * bell(0).conj().T
        rho2 = bell(1) * bell(1).conj().T
        states = [rho1, rho2]
        state_distinguishability(states, [1, 2, 3])


def test_invalid_state_distinguishability_states():
    """Invalid number of states for state distinguishability."""
    with np.testing.assert_raises(ValueError):
        states = []
        state_distinguishability(states)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Gell-Mann matrices."""
from scipy.sparse import csr_matrix

import numpy as np


def gell_mann(ind: int, is_sparse: bool = False) -> np.ndarray:
    r"""
    Produce a Gell-Mann operator [WikGM]_.

    Generates the 3-by-3 Gell-Mann matrix indicated by the value of `ind`.
    `ind = 0` gives the identity matrix, while values 1 through 8 each indicate
    one of the other 8 Gell-Mann matrices.

    The 9 Gell-Mann matrices are defined as follows:

    .. math::
        \begin{equation}
            \begin{aligned}
                \lambda_0 = \begin{pmatrix}
                                1 & 0 & 0 \\
                                0 & 1 & 0 \\
                                0 & 0 & 1
                            \end{pmatrix}, \quad
                \lambda_1 = \begin{pmatrix}
                                0 & 1 & 0 \\
                                1 & 0 & 0 \\
                                0 & 0 & 0
                            \end{pmatrix}, \quad &
                \lambda_2 = \begin{pmatrix}
                                0 & -i & 0 \\
                                i & 0 & 0 \\
                                0 & 0 & 0
                            \end{pmatrix},  \\
                \lambda_3 = \begin{pmatrix}
                                1 & 0 & 0 \\
                                0 & -1 & 0 \\
                                0 & 0 & 0
                            \end{pmatrix}, \quad
                \lambda_4 = \begin{pmatrix}
                                0 & 0 & 1 \\
                                0 & 0 & 0 \\
                                1 & 0 & 0
                            \end{pmatrix}, \quad &
                \lambda_5 = \begin{pmatrix}
                                0 & 0 & -i \\
                                0 & 0 & 0 \\
                                i & 0 & 0
                            \end{pmatrix},  \\
                \lambda_6 = \begin{pmatrix}
                                0 & 0 & 0 \\
                                0 & 0 & 1 \\
                                0 & 1 & 0
                            \end{pmatrix}, \quad
                \lambda_7 = \begin{pmatrix}
                                0 & 0 & 0 \\
                                0 & 0 & -i \\
                                0 & i & 0
                            \end{pmatrix}, \quad &
                \lambda_8 = \frac{1}{\sqrt{3}} \begin{pmatrix}
                                                    1 & 0 & 0 \\
                                                    0 & 1 & 0 \\
                                                    0 & 0 & -2
                                                \end{pmatrix}.
                \end{aligned}
            \end{equation}

    Examples
    ==========

    The Gell-Mann matrix generated from :math:`idx = 2` yields the following
    matrix:

    .. math::

        \lambda_2 = \begin{pmatrix}
                            0 & -i & 0 \\
                            i & 0 & 0 \\
                            0 & 0 & 0
                    \end{pmatrix}

    >>> from toqito.matrices import gell_mann
    >>> gell_mann(2)
    [[ 0.+0.j, -0.-1.j,  0.+0.j],
     [ 0.+1.j,  0.+0.j,  0.+0.j],
     [ 0.+0.j,  0.+0.j,  0.+0.j]]

    References
    ==========
    .. [WikGM] Wikipedia: Gell-Mann matrices,
        https://en.wikipedia.org/wiki/Gell-Mann_matrices

    :param ind: An integer between 0 and 8 (inclusive).
    :param is_sparse: Boolean to determine whether matrix is sparse.
    """
    if ind == 0:
        gm_op = np.identity(3)
    elif ind == 1:
        gm_op = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]])
    elif ind == 2:
        gm_op = np.array([[0, -1j, 0], [1j, 0, 0], [0, 0, 0]])
    elif ind == 3:
        gm_op = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 0]])
    elif ind == 4:
        gm_op = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]])
    elif ind == 5:
        gm_op = np.array([[0, 0, -1j], [0, 0, 0], [1j, 0, 0]])
    elif ind == 6:
        gm_op = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]])
    elif ind == 7:
        gm_op = np.array([[0, 0, 0], [0, 0, -1j], [0, 1j, 0]])
    elif ind == 8:
        gm_op = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -2]]) / np.sqrt(3)
    else:
        raise ValueError("Gell-Mann index values can only be values from 0 to " "8 (inclusive).")

    if is_sparse:
        gm_op = csr_matrix(gm_op)

    return gm_op



"""Test bell."""
import numpy as np

from toqito.states import basis
from toqito.states import bell


def test_bell_0():
    """Generate the Bell state: `1/sqrt(2) * (|00> + |11>)`."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = 1 / np.sqrt(2) * (np.kron(e_0, e_0) + np.kron(e_1, e_1))

    res = bell(0)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_bell_1():
    """Generates the Bell state: `1/sqrt(2) * (|00> - |11>)`."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = 1 / np.sqrt(2) * (np.kron(e_0, e_0) - np.kron(e_1, e_1))

    res = bell(1)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_bell_2():
    """Generates the Bell state: `1/sqrt(2) * (|01> + |10>)`."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = 1 / np.sqrt(2) * (np.kron(e_0, e_1) + np.kron(e_1, e_0))

    res = bell(2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_bell_3():
    """Generates the Bell state: `1/sqrt(2) * (|01> - |10>)`."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = 1 / np.sqrt(2) * (np.kron(e_0, e_1) - np.kron(e_1, e_0))

    res = bell(3)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_bell_invalid():
    """Ensures that an integer above 3 is error-checked."""
    with np.testing.assert_raises(ValueError):
        bell(4)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test max_mixed."""
import numpy as np

from toqito.states import max_mixed


def test_max_mixed_dim_2_full():
    """Generate full 2-dimensional maximally mixed state."""
    expected_res = 1 / 2 * np.array([[1, 0], [0, 1]])
    res = max_mixed(2, is_sparse=False)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_max_mixed_dim_2_sparse():
    """Generate sparse 2-dimensional maximally mixed state."""
    expected_res = 1 / 2 * np.array([[1, 0], [0, 1]])
    res = max_mixed(2, is_sparse=True)

    bool_mat = np.isclose(res.toarray(), expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Generalized Gell-Mann matrices."""
from typing import Union
from scipy import sparse

import numpy as np


def gen_gell_mann(
    ind_1: int, ind_2: int, dim: int, is_sparse: bool = False
) -> Union[np.ndarray, sparse.lil_matrix]:
    r"""
    Produce a generalized Gell-Mann operator [WikGM2]_.

    Construct a `dim`-by-`dim` Hermitian operator. These matrices span the
    entire space of `dim`-by-`dim` matrices as `ind_1` and `ind_2` range from 0
    to `dim-1`, inclusive, and they generalize the Pauli operators when `dim =
    2` and the Gell-Mann operators when `dim = 3`.

    Examples
    ==========

    The generalized Gell-Mann matrix for `ind_1 = 0`, `ind_2 = 1` and `dim = 2`
    is given as

    .. math::
        G_{0, 1, 2} = \begin{pmatrix}
                         0 & 1 \\
                         1 & 0
                      \end{pmatrix}.

    This can be obtained in `toqito` as follows.

    >>> from toqito.matrices import gen_gell_mann
    >>> gen_gell_mann(0, 1, 2)
    [[0., 1.],
     [1., 0.]])

    The generalized Gell-Mann matrix `ind_1 = 2`, `ind_2 = 3`, and `dim = 4` is
    given as

    .. math::
        G_{2, 3, 4} = \begin{pmatrix}
                        0 & 0 & 0 & 0 \\
                        0 & 0 & 0 & 0 \\
                        0 & 0 & 0 & 1 \\
                        0 & 0 & 1 & 0
                      \end{pmatrix}.

    This can be obtained in `toqito` as follows.

    >>> from toqito.matrices import gen_gell_mann
    >>> gen_gell_mann(2, 3, 4)
    [[0., 0., 0., 0.],
     [0., 0., 0., 0.],
     [0., 0., 0., 1.],
     [0., 0., 1., 0.]])

    References
    ==========
    .. [WikGM2] Wikipedia: Gell-Mann matrices,
        https://en.wikipedia.org/wiki/Gell-Mann_matrices

    :param ind_1: A non-negative integer from 0 to `dim-1` (inclusive).
    :param ind_2: A non-negative integer from 0 to `dim-1` (inclusive).
    :param dim: The dimension of the Gell-Mann operator.
    :param is_sparse: If set to `True`, the returned Gell-Mann operator is a
                      sparse lil_matrix and if set to `False`, the returned
                      Gell-Mann operator is a dense numpy array.
    :return: The generalized Gell-Mann operator.
    """
    if ind_1 == ind_2:
        if ind_1 == 0:
            gm_op = sparse.eye(dim)
        else:
            scalar = np.sqrt(2 / (ind_1 * (ind_1 + 1)))
            diag = np.ones((ind_1, 1))
            diag = np.append(diag, -ind_1)
            diag = scalar * np.append(diag, np.zeros((dim - ind_1 - 1, 1)))

            gm_op = sparse.lil_matrix((dim, dim))
            gm_op.setdiag(diag)

    else:
        e_mat = sparse.lil_matrix((dim, dim))
        e_mat[ind_1, ind_2] = 1
        if ind_1 < ind_2:
            gm_op = e_mat + e_mat.conj().T
        else:
            gm_op = 1j * e_mat - 1j * e_mat.conj().T

    if not is_sparse:
        return gm_op.todense()
    return gm_op



"""Trace distance metric."""
import numpy as np

from toqito.matrix_props import is_density
from toqito.state_metrics import trace_norm


def trace_distance(rho: np.ndarray, sigma: np.ndarray) -> float:
    r"""
    Compute the trace distance between density operators `rho` and `sigma`.

    The trace distance between :math:`\rho` and :math:`\sigma` is defined as

    .. math::
        \delta(\rho, \sigma) = \frac{1}{2} \left( \text{Tr}(\left| \rho - \sigma
         \right| \right).

    More information on the trace distance can be found in [WIKTD]_.

    Examples
    ==========

    Consider the following Bell state

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \in \mathcal{X}.

    The corresponding density matrix of :math:`u` may be calculated by:

    .. math::
        \rho = u u^* = \begin{pmatrix}
                         1 & 0 & 0 & 1 \\
                         0 & 0 & 0 & 0 \\
                         0 & 0 & 0 & 0 \\
                         1 & 0 & 0 & 1
                       \end{pmatrix} \text{D}(\mathcal{X}).

    The trace distance between :math:`\rho` and another state :math:`\sigma` is equal to :math:`0`
    if any only if :math:`\rho = \sigma`. We can check this using the :code:`toqito` package.

    >>> from toqito.states import bell
    >>> from toqito.state_metrics import trace_norm
    >>> rho = bell(0) * bell(0).conj().T
    >>> sigma = rho
    >>> trace_distance(rho, sigma)
    0.0

    References
    ==========
    .. [WIKTD] Quantiki: Trace distance
            https://www.quantiki.org/wiki/trace-distance

    :param rho: An input matrix.
    :param sigma: An input matrix.
    :return: The trace distance between :code:`rho` and :code:`sigma`.
    """
    if not is_density(rho) or not is_density(sigma):
        raise ValueError("Trace distance is only defined for density " "operators.")
    return trace_norm(np.abs(rho - sigma)) / 2



"""Test concurrence."""
import numpy as np

from toqito.state_props import concurrence
from toqito.states import basis


def test_concurrence_entangled():
    """The concurrence on maximally entangled Bell state."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00, e_11 = np.kron(e_0, e_0), np.kron(e_1, e_1)

    u_vec = 1 / np.sqrt(2) * (e_00 + e_11)
    rho = u_vec * u_vec.conj().T

    res = concurrence(rho)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_concurrence_separable():
    """The concurrence of a product state is zero."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    v_vec = np.kron(e_0, e_1)
    sigma = v_vec * v_vec.conj().T

    res = concurrence(sigma)
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_concurrence_invalid_dim():
    """Tests for invalid dimension inputs."""
    with np.testing.assert_raises(ValueError):
        rho = np.identity(5)
        concurrence(rho)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test domino."""
import numpy as np

from toqito.states import basis
from toqito.states import domino


def test_domino_0():
    """Domino with index = 0."""
    e_1 = basis(3, 1)
    expected_res = np.kron(e_1, e_1)
    res = domino(0)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_1():
    """Domino with index = 1."""
    e_0, e_1 = basis(3, 0), basis(3, 1)
    expected_res = np.kron(e_0, 1 / np.sqrt(2) * (e_0 + e_1))
    res = domino(1)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_2():
    """Domino with index = 2."""
    e_0, e_1 = basis(3, 0), basis(3, 1)
    expected_res = np.kron(e_0, 1 / np.sqrt(2) * (e_0 - e_1))
    res = domino(2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_3():
    """Domino with index = 3."""
    e_1, e_2 = basis(3, 1), basis(3, 2)
    expected_res = np.kron(e_2, 1 / np.sqrt(2) * (e_1 + e_2))
    res = domino(3)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_4():
    """Domino with index = 4."""
    e_1, e_2 = basis(3, 1), basis(3, 2)
    expected_res = np.kron(e_2, 1 / np.sqrt(2) * (e_1 - e_2))
    res = domino(4)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_5():
    """Domino with index = 5."""
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    expected_res = np.kron(1 / np.sqrt(2) * (e_1 + e_2), e_0)
    res = domino(5)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_6():
    """Domino with index = 6."""
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    expected_res = np.kron(1 / np.sqrt(2) * (e_1 - e_2), e_0)
    res = domino(6)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_7():
    """Domino with index = 7."""
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    expected_res = np.kron(1 / np.sqrt(2) * (e_0 + e_1), e_2)
    res = domino(7)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_8():
    """Domino with index = 8."""
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    expected_res = np.kron(1 / np.sqrt(2) * (e_0 - e_1), e_2)
    res = domino(8)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_domino_invalid_index():
    """Tests for invalid index input."""
    with np.testing.assert_raises(ValueError):
        domino(9)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Compute the Choi matrix of a list of Kraus operators."""
from typing import List
import numpy as np

from toqito.states import max_entangled
from toqito.channel_ops import partial_channel


def kraus_to_choi(kraus_ops: List[List[np.ndarray]], sys: int = 2) -> np.ndarray:
    r"""
    Compute the Choi matrix of a list of Kraus operators [WatKraus18]_.

    The Choi matrix of the list of Kraus operators, :code:`kraus_ops`. The default convention is
    that the Choi matrix is the result of applying the map to the second subsystem of the
    standard maximally entangled (unnormalized) state. The Kraus operators are expected to be
    input as a list of numpy arrays.

    This function was adapted from the QETLAB package.

    Examples
    ==========

    The transpose map:

    The Choi matrix of the transpose map is the swap operator.

    >>> import numpy as np
    >>> from toqito.channel_ops import kraus_to_choi
    >>> kraus_1 = np.array([[1, 0], [0, 0]])
    >>> kraus_2 = np.array([[1, 0], [0, 0]]).conj().T
    >>> kraus_3 = np.array([[0, 1], [0, 0]])
    >>> kraus_4 = np.array([[0, 1], [0, 0]]).conj().T
    >>> kraus_5 = np.array([[0, 0], [1, 0]])
    >>> kraus_6 = np.array([[0, 0], [1, 0]]).conj().T
    >>> kraus_7 = np.array([[0, 0], [0, 1]])
    >>> kraus_8 = np.array([[0, 0], [0, 1]]).conj().T
    >>>
    >>> kraus_ops = [
    >>>     [kraus_1, kraus_2],
    >>>     [kraus_3, kraus_4],
    >>>     [kraus_5, kraus_6],
    >>>     [kraus_7, kraus_8],
    >>> ]
    >>> kraus_to_choi(kraus_ops)
    [[1. 0. 0. 0.]
     [0. 0. 1. 0.]
     [0. 1. 0. 0.]
     [0. 0. 0. 1.]]

    See Also
    ========
    choi_to_kraus

    References
    ==========
    .. [WatKraus18] Watrous, John.
        "The theory of quantum information."
        Section: "Kraus representations".
        Cambridge University Press, 2018.

    :param kraus_ops: A list of Kraus operators.
    :param sys: The dimension of the system (default is 2).
    :return: The corresponding Choi matrix of the provided Kraus operators.
    """
    dim_op_1 = kraus_ops[0][0].shape[0]
    dim_op_2 = kraus_ops[0][0].shape[1]

    choi_mat = partial_channel(
        max_entangled(dim_op_1, False, False) * max_entangled(dim_op_2, False, False).conj().T,
        kraus_ops,
        sys,
        np.array([[dim_op_1, dim_op_1], [dim_op_2, dim_op_2]]),
    )

    return choi_mat



"""Quantum states."""
from toqito.states.basis import basis
from toqito.states.bell import bell
from toqito.states.chessboard import chessboard
from toqito.states.domino import domino
from toqito.states.gen_bell import gen_bell
from toqito.states.ghz import ghz
from toqito.states.gisin import gisin
from toqito.states.horodecki import horodecki
from toqito.states.max_entangled import max_entangled
from toqito.states.max_mixed import max_mixed
from toqito.states.isotropic import isotropic
from toqito.states.tile import tile
from toqito.states.w_state import w_state
from toqito.states.werner import werner
from toqito.states.breuer import breuer
from toqito.states.brauer import brauer



"""Test pauli."""
import numpy as np

from scipy.sparse import issparse

from toqito.matrices import pauli


def test_pauli_str_sparse():
    """Pauli-I operator with argument "I"."""
    res = pauli("I", True)
    np.testing.assert_equal(issparse(res), True)


def test_pauli_int_sparse():
    """Pauli-I operator with argument "I"."""
    res = pauli(0, True)
    np.testing.assert_equal(issparse(res), True)


def test_pauli_i():
    """Pauli-I operator with argument "I"."""
    expected_res = np.array([[1, 0], [0, 1]])
    res = pauli("I")

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pauli_x():
    """Pauli-X operator with argument "X"."""
    expected_res = np.array([[0, 1], [1, 0]])
    res = pauli("X")

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pauli_1():
    """Pauli-X operator with argument 1."""
    expected_res = np.array([[0, 1], [1, 0]])
    res = pauli(1)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pauli_y():
    """Pauli-Y operator with argument "Y"."""
    expected_res = np.array([[0, -1j], [1j, 0]])
    res = pauli("Y")

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pauli_2():
    """Pauli-Y operator with argument 2."""
    expected_res = np.array([[0, -1j], [1j, 0]])
    res = pauli(2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pauli_z():
    """Pauli-Z operator with argument "Z"."""
    expected_res = np.array([[1, 0], [0, -1]])
    res = pauli("Z")

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pauli_3():
    """Pauli-Z operator with argument 3."""
    expected_res = np.array([[1, 0], [0, -1]])
    res = pauli(3)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pauli_int_list():
    """Test with list of Paulis of ints."""
    expected_res = np.array([[0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0], [1, 0, 0, 0]])
    res = pauli([1, 1])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_pauli_str_list():
    """Test with list of Paulis of str."""
    expected_res = np.array([[0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0], [1, 0, 0, 0]])
    res = pauli(["x", "x"])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Permutation and combinatorial functions."""
from toqito.perms.unique_perms import unique_perms
from toqito.perms.perfect_matchings import perfect_matchings
from toqito.perms.perm_sign import perm_sign
from toqito.perms.permute_systems import permute_systems
from toqito.perms.permutation_operator import permutation_operator
from toqito.perms.swap import swap
from toqito.perms.swap_operator import swap_operator
from toqito.perms.antisymmetric_projection import antisymmetric_projection
from toqito.perms.symmetric_projection import symmetric_projection



"""Antisymmetric projection operator."""
from itertools import permutations
from scipy import linalg, sparse

import numpy as np

from toqito.perms import permutation_operator
from toqito.perms import perm_sign


def antisymmetric_projection(
    dim: int, p_param: int = 2, partial: bool = False
) -> sparse.lil_matrix:
    r"""
    Produce the projection onto the antisymmetric subspace [WikAsym]_.

    Produces the orthogonal projection onto the anti-symmetric subspace of :code:`p_param` copies of
    :code:`dim`-dimensional space. If :code:`partial = True`, then the antisymmetric projection (PA)
    isn't the orthogonal projection itself, but rather a matrix whose columns form an orthonormal
    basis for the symmetric subspace (and hence the PA * PA' is the orthogonal projection onto the
    symmetric subspace.)

    Examples
    ==========

    The :math:`2`-dimensional antisymmetric projection with :math:`p=1` is given as
    :math:`2`-by-:math:`2` identity matrix

    .. math::
        A_{2,1} =
        \begin{pmatrix}
            1 & 0 \\
            0 & 1
        \end{pmatrix}.

    Using :code:`toqito`, we can see this gives the proper result.

    >>> from toqito.perms import antisymmetric_projection
    >>> antisymmetric_projection(2, 1).todense()
    [[1., 0.],
     [0., 1.]]

    When the :math:`p` value is greater than the dimension of the antisymmetric projection, this
    just gives the matrix consisting of all zero entries. For instance, when :math:`d = 2` and
    :math:`p = 3` we have that

    .. math::
        A_{2, 3} =
        \begin{pmatrix}
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\
            0 & 0 & 0 & 0 & 0 & 0 & 0 & 0
        \end{pmatrix}.

    Using :code:`toqito` we can see this gives the proper result.

    >>> from toqito.perms import antisymmetric_projection
    >>> antisymmetric_projection(2, 3).todense()
    [[0., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 0.]]

    References
    ==========
    .. [WikAsym] Wikipedia: Anti-symmetric operator
        https://en.wikipedia.org/wiki/Anti-symmetric_operator

    :param dim: The dimension of the local systems.
    :param p_param: Default value of 2.
    :param partial: Default value of 0.
    :return: Projection onto the antisymmetric subspace.
    """
    dimp = dim ** p_param

    if p_param == 1:
        return sparse.eye(dim)
    # The antisymmetric subspace is empty if `dim < p`.
    if dim < p_param:
        return sparse.lil_matrix((dimp, dimp * (1 - partial)))

    p_list = np.array(list(permutations(np.arange(1, p_param + 1))))
    p_fac = p_list.shape[0]

    anti_proj = sparse.lil_matrix((dimp, dimp))
    for j in range(p_fac):
        anti_proj += perm_sign(p_list[j, :]) * permutation_operator(
            dim * np.ones(p_param), p_list[j, :], False, True
        )
    anti_proj = anti_proj / p_fac

    if partial:
        anti_proj = anti_proj.todense()
        anti_proj = sparse.lil_matrix(linalg.orth(anti_proj))
    return anti_proj



"""Test unvec."""
import numpy as np

from toqito.matrix_ops import unvec


def test_unvec():
    """Test standard unvec operation on a vector."""
    expected_res = np.array([[1, 2], [3, 4]])

    test_input_vec = np.array([1, 3, 2, 4])

    res = unvec(test_input_vec)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_unvec_custom_dim():
    """Test standard unvec operation on a vector with custom dimension."""
    expected_res = np.array([[1], [3], [2], [4]])

    test_input_vec = np.array([1, 3, 2, 4])

    res = unvec(test_input_vec, [1, 4])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_projection."""
import numpy as np

from toqito.matrix_props import is_projection


def test_is_projection():
    """Check that non-Hermitian projection matrix returns False."""
    mat = np.array([[0, 1], [0, 1]])
    np.testing.assert_equal(is_projection(mat), False)


def test_is_projection_2():
    """Check that projection matrix returns True."""
    mat = np.array([[1, 0], [0, 1]])
    np.testing.assert_equal(is_projection(mat), True)


def test_is_not_pd_non_projection():
    """Check that non-projection matrix returns False."""
    mat = np.array([[-1, -1], [-1, -1]])
    np.testing.assert_equal(is_projection(mat), False)


def test_is_projection_not_square():
    """Input must be a square matrix."""
    mat = np.array([[-1, 1, 1], [1, 2, 3]])
    np.testing.assert_equal(is_projection(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test isotropic."""
import numpy as np

from toqito.states import isotropic


def test_isotropic_qutrit():
    """Generate a qutrit isotropic state with `alpha` = 1/2."""
    res = isotropic(3, 1 / 2)

    np.testing.assert_equal(np.isclose(res[0, 0], 2 / 9), True)
    np.testing.assert_equal(np.isclose(res[4, 4], 2 / 9), True)
    np.testing.assert_equal(np.isclose(res[8, 8], 2 / 9), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test gen_pauli."""
import numpy as np

from toqito.matrices import gen_pauli


def test_gen_pauli_1_0_2():
    """Generalized Pauli operator for k_1 = 1, k_2 = 0, and dim = 2."""
    dim = 2
    k_1 = 1
    k_2 = 0

    # Pauli-X operator.
    expected_res = np.array([[0, 1], [1, 0]])
    res = gen_pauli(k_1, k_2, dim)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gen_pauli_0_1_2():
    """Generalized Pauli operator for k_1 = 0, k_2 = 1, and dim = 2."""
    dim = 2
    k_1 = 0
    k_2 = 1

    # Pauli-Z operator.
    expected_res = np.array([[1, 0], [0, -1]])
    res = gen_pauli(k_1, k_2, dim)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gen_pauli_1_1_2():
    """Generalized Pauli operator for k_1 = 1, k_2 = 1, and dim = 2."""
    dim = 2
    k_1 = 1
    k_2 = 1

    # Pauli-Y operator.
    expected_res = np.array([[0, -1], [1, 0]])
    res = gen_pauli(k_1, k_2, dim)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Compute the l1-norm of coherence of a quantum state."""
import numpy as np
from toqito.state_ops import pure_to_mixed


def l1_norm_coherence(rho: np.ndarray) -> float:
    r"""
    Compute the l1-norm of coherence of a quantum state [RPWL17]_.

    The :math:`\ell_1`-norm of coherence of a quantum state :math:`\rho` is
    defined as

    .. math::
        C_{\ell_1}(\rho) = \sum_{i \not= j} \left|\rho_{i,j}\right|,

    where :math:`\rho_{i,j}` is the :math:`(i,j)^{th}`-entry of :math:`\rho`
    in the standard basis.

    The :math:`\ell_1`-norm of coherence is the sum of the absolute values of
    the sum of the absolute values of the off-diagonal entries of the density
    matrix :code:`rho` in the standard basis.

    This function was adapted from QETLAB.

    Examples
    ========

    The largest possible value of the :math:`\ell_1`-norm of coherence on
    :math:`d`-dimensional states is :math:`d-1`, and is attained exactly by
    the "maximally coherent states": pure states whose entries all have the
    same absolute value.

    >>> from toqito.state_props import l1_norm_coherence
    >>> import numpy as np
    >>>
    >>> # Maximally coherent state.
    >>> v = np.ones((3,1))/np.sqrt(3)
    >>> l1_norm_coherence(v)
    2

    References
    ==========
    .. [RPWL17] "Logarithmic coherence: Operational interpretation of 1-norm coherence"
        Rana, Swapan, Parashar, Preeti, Winter, Andreas, Lewenstein, Maciej
        Physical Review A 96.5 (2017): 052336.
        https://arxiv.org/pdf/1612.09234.pdf

    :param rho: A matrix or vector.
    :return: The l1-norm coherence of :code:`rho`.
    """
    rho = pure_to_mixed(rho)
    return np.sum(np.sum(np.abs(rho))) - np.trace(rho)



"""Fourier matrix."""
import numpy as np


def fourier(dim: int) -> np.ndarray:
    r"""
    Generate the Fourier transform matrix [WikDFT]_.

    Generates the `dim`-by-`dim` unitary matrix that implements the quantum
    Fourier transform.

    The Fourier matrix is defined as:

    .. math::
        W_N = \frac{1}{N}
        \begin{pmatrix}
            1 & 1 & 1 & 1 & \ldots & 1 \\
            1 & \omega & \omega^2 & \omega^3 & \ldots & \omega^{N-1} \\
            1 & \omega^2 & \omega^4 & \omega^6 & \ldots & \omega^{2(N-1)} \\
            1 & \omega^3 & \omega^6 & \omega^9 & \ldots & \omega^{3(N-1)} \\
            \vdots & \vdots & \vdots & \vdots & \ddots & \vdots \\
            1 & \omega^{N-1} & \omega^{2(N-1)} & \omega^{3(N-1)} &
            \ldots & \omega^{3(N-1)}
        \end{pmatrix}

    Examples
    ==========

    The Fourier matrix generated from :math:`d = 3` yields the following
    matrix:

    .. math::
        W_3 = \frac{1}{3}
        \begin{pmatrix}
            1 & 1 & 1 \\
            0 & \omega & \omega^2 \\
            1 & \omega^2 & \omega^4
        \end{pmatrix}

    >>> from toqito.matrices import fourier
    >>> fourier(3)
    [[ 0.57735027+0.j ,  0.57735027+0.j ,  0.57735027+0.j ],
     [ 0.57735027+0.j , -0.28867513+0.5j, -0.28867513-0.5j],
     [ 0.57735027+0.j , -0.28867513-0.5j, -0.28867513+0.5j]]

    References
    ==========
    .. [WikDFT] Wikipedia: DFT matrix,
        https://en.wikipedia.org/wiki/DFT_matrix

    :param dim: The size of the Fourier matrix.
    :return: The Fourier matrix of dimension `dim`.
    """
    # Primitive root of unity.
    root_unity = np.exp(2 * 1j * np.pi / dim)
    entry_1 = np.arange(0, dim)[:, None]
    entry_2 = np.arange(0, dim)
    return np.power(root_unity, entry_1 * entry_2) / np.sqrt(dim)



"""Tests for trace_distance."""
import numpy as np

from toqito.state_metrics import trace_distance
from toqito.states import basis


def test_trace_distance_same_state():
    r"""Test that: :math:`T(\rho, \sigma) = 0` iff `\rho = \sigma`."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00 = np.kron(e_0, e_0)
    e_11 = np.kron(e_1, e_1)

    u_vec = 1 / np.sqrt(2) * (e_00 + e_11)
    rho = u_vec * u_vec.conj().T
    sigma = rho

    res = trace_distance(rho, sigma)

    np.testing.assert_equal(np.isclose(res, 0), True)


def test_trace_distance_non_density_matrix():
    r"""Test trace distance on non-density matrix."""
    rho = np.array([[1, 2], [3, 4]])
    sigma = np.array([[5, 6], [7, 8]])

    with np.testing.assert_raises(ValueError):
        trace_distance(rho, sigma)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for helstrom_holevo."""
import numpy as np

from toqito.state_metrics import helstrom_holevo
from toqito.states import basis


def test_helstrom_holevo_same_state():
    """Test Helstrom-Holevo distance on same state."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00 = np.kron(e_0, e_0)
    e_11 = np.kron(e_1, e_1)

    u_vec = 1 / np.sqrt(2) * (e_00 + e_11)
    sigma_0 = u_vec * u_vec.conj().T
    sigma_1 = sigma_0

    res = helstrom_holevo(sigma_0, sigma_1)

    np.testing.assert_equal(np.isclose(res, 1 / 2), True)


def test_helstrom_holevo_non_density_matrix():
    """Test Helstrom-Holevo distance on non-density matrix."""
    rho = np.array([[1, 2], [3, 4]])
    sigma = np.array([[5, 6], [7, 8]])

    with np.testing.assert_raises(ValueError):
        helstrom_holevo(rho, sigma)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tile state."""
import numpy as np

from toqito.states import basis


def tile(idx: int) -> np.ndarray:
    r"""
    Produce a Tile state [UPBTile99]_.

    The Tile states constitute five states on 3-by-3 dimensional space that form a UPB
    (unextendible product basis).

    Returns one of the following five tile states depending on the value of :code:`idx`:

    .. math::
        \begin{equation}
            \begin{aligned}
                |\psi_0 \rangle = \frac{1}{\sqrt{2}} |0 \rangle
                \left(|0\rangle - |1\rangle \right),
                \qquad &
                |\psi_1\rangle = \frac{1}{\sqrt{2}}
                \left(|0\rangle - |1\rangle \right) |2\rangle, \\
                |\psi_2\rangle = \frac{1}{\sqrt{2}} |2\rangle
                \left(|1\rangle - |2\rangle \right),
                \qquad &
                |\psi_3\rangle = \frac{1}{\sqrt{2}}
                \left(|1\rangle - |2\rangle \right) |0\rangle, \\
                \qquad &
                |\psi_4\rangle = \frac{1}{3}
                \left(|0\rangle + |1\rangle + |2\rangle)\right)
                \left(|0\rangle + |1\rangle + |2\rangle \right).
            \end{aligned}
        \end{equation}

    Examples
    ==========

    When :code:`idx = 0`, this produces the following tile state

    .. math::
        \frac{1}{\sqrt{2}} |0\rangle \left( |0\rangle - |1\rangle \right).

    Using :code:`toqito`, we can see that this yields the proper state.

    >>> from toqito.states import tile
    >>> import numpy as np
    >>> tile(0)
    [[ 0.70710678]
     [-0.        ]
     [ 0.        ]]

    References
    ==========
    .. [UPBTile99] Bennett, Charles H., et al.
        "Unextendible product bases and bound entanglement."
        Physical Review Letters 82.26 (1999): 5385.
        https://arxiv.org/abs/quant-ph/9808030

    :param idx: A parameter in [0, 1, 2, 3, 4]
    :return: Tile state.
    """
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    if idx == 0:
        return 1 / np.sqrt(2) * np.kron(e_0, (e_0 - e_1))
    if idx == 1:
        return 1 / np.sqrt(2) * np.kron((e_0 - e_1), e_2)
    if idx == 2:
        return 1 / np.sqrt(2) * np.kron(e_2, (e_1 - e_2))
    if idx == 3:
        return 1 / np.sqrt(2) * np.kron((e_1 - e_2), e_0)
    if idx == 4:
        return 1 / 3 * np.kron((e_0 + e_1 + e_2), (e_0 + e_1 + e_2))
    raise ValueError("Invalid integer value for Tile state.")



"""Is channel trace-preserving."""
from typing import List, Union

import numpy as np

from toqito.matrix_props import is_identity
from toqito.channels import partial_trace


def is_trace_preserving(
    phi: Union[np.ndarray, List[List[np.ndarray]]],
    rtol: float = 1e-05,
    atol: float = 1e-08,
    sys: Union[int, List[int]] = 2,
    dim: Union[List[int], np.ndarray] = None,
) -> bool:
    r"""
    Determine whether the given channel is trace-preserving [WatTP18]_.

    A map :math:`\Phi \in \text{T} \left(\mathcal{X}, \mathcal{Y} \right)` is
    *trace-preserving* if it holds that

    .. math::
        \text{Tr} \left( \Phi(X) \right) = \text{Tr}\left( X \right)

    for every operator :math:`X \in \text{L}(\mathcal{X})`.

    Given the corresponding Choi matrix of the channel, a neccessary and sufficient condition is

    .. math::
        \text{Tr}_{\mathcal{Y}} \left( J(\Phi) \right) = \mathbb{I}_{\mathcal{X}}

    In case :code:`sys` is not specified, the default convention is that the Choi matrix
    is the result of applying the map to the second subsystem of the standard maximally
    entangled (unnormalized) state.

    The dimensions of the subsystems are given by the vector :code:`dim`. By default,
    both subsystems have equal dimension.

    Alternatively, given a list of Kraus operators, a neccessary and sufficient condition is

    .. math::
        \sum_{a \in \Sigma} A_a^* B_a = \mathbb{I}_{\mathcal{X}}

    Examples
    ==========

    The map :math:`\Phi` defined as

    .. math::
        \Phi(X) = X - U X U^*

    is not trace-preserving, where

    .. math::
        U = \frac{1}{\sqrt{2}}
        \begin{pmatrix}
            1 & 1 \\
            -1 & 1
        \end{pmatrix}.

    >>> import numpy as np
    >>> from toqito.channel_props import is_trace_preserving
    >>> unitary_mat = np.array([[1, 1], [-1, 1]]) / np.sqrt(2)
    >>> kraus_ops = [[np.identity(2), np.identity(2)], [unitary_mat, -unitary_mat]]
    >>> is_trace_preserving(kraus_ops)
    False

    As another example, the depolarizing channel is trace-preserving.

    >>> from toqito.channels import depolarizing
    >>> from toqito.channel_props import is_trace_preserving
    >>> choi_mat = depolarizing(2)
    >>> is_trace_preserving(choi_mat)
    True

    References
    ==========
    .. [WatTP18] Watrous, John.
        "The theory of quantum information."
        Section: "Linear maps of square operators".
        Cambridge University Press, 2018.

    :param phi: The channel provided as either a Choi matrix or a list of Kraus operators.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :param sys: Scalar or vector specifying the size of the subsystems.
    :param dim: Dimension of the subsystems. If :code:`None`, all dimensions are assumed to be
                equal.
    :return: True if the channel is trace-preserving, and False otherwise.
    """
    # If the variable `phi` is provided as a list, we assume this is a list
    # of Kraus operators.
    if isinstance(phi, list):
        phi_l = [A for A, _ in phi]
        phi_r = [B for _, B in phi]

        k_l = np.concatenate(phi_l, axis=0)
        k_r = np.concatenate(phi_r, axis=0)

        mat = k_l.conj().T @ k_r
    else:
        mat = partial_trace(input_mat=phi, sys=sys, dim=dim)
    return is_identity(mat, rtol=rtol, atol=atol)



"""Is matrix a square matrix."""
import numpy as np


def is_square(mat: np.ndarray) -> bool:
    r"""
    Determine if a matrix is square [WikSquare]_.

    A matrix is square if the dimensions of the rows and columns are equivalent.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6 \\
                7 & 8 & 9
            \end{pmatrix}

    our function indicates that this is indeed a square matrix.

    >>> from toqito.matrix_props import is_square
    >>> import numpy as np
    >>> A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> is_square(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6
            \end{pmatrix}

    is not square.

    >>> from toqito.matrix_props import is_square
    >>> import numpy as np
    >>> B = np.array([[1, 2, 3], [4, 5, 6]])
    >>> is_square(B)
    False

    References
    ==========
    .. [WikSquare] Wikipedia: Square matrix.
        https://en.wikipedia.org/wiki/Square_matrix

    :param mat: The matrix to check.
    :return: Returns :code:`True` if the matrix is square and :code:`False` otherwise.
    """
    if len(mat.shape) != 2:
        raise ValueError("The variable is not a matrix.")
    return mat.shape[0] == mat.shape[1]



"""Test cvx_kron."""
import cvxpy
import numpy as np

from toqito.helper import cvx_kron


def test_cvx_kron():
    """Ensure return type is CVX object."""
    rho = cvxpy.Variable((4, 4))
    res_mat = cvx_kron(rho, np.identity(4))
    np.testing.assert_equal(isinstance(res_mat, cvxpy.atoms.affine.vstack.Vstack), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""The dephasing channel."""
import numpy as np

from toqito.states import max_entangled


def dephasing(dim: int, param_p: float = 0) -> np.ndarray:
    r"""
    Produce the partially dephasing channel [WatDeph18]_.

    The Choi matrix of the completely dephasing channel that acts on :code:`dim`-by-:code:`dim`
    matrices.

    Let :math:`\Sigma` be an alphabet and let :math:`\mathcal{X} = \mathbb{C}^{\Sigma}`. The map
    :math:`\Delta \in \text{T}(\mathcal{X})` defined as

    .. math::
        \Delta(X) = \sum_{a \in \Sigma} X(a, a) E_{a,a}

    for every :math:`X \in \text{L}(\mathcal{X})` is defined as the *completely dephasing channel*.

    Examples
    ==========

    The completely dephasing channel maps kills everything off the diagonal. Consider the
    following matrix

    .. math::
        \rho = \begin{pmatrix}
                   1 & 2 & 3 & 4 \\
                   5 & 6 & 7 & 8 \\
                   9 & 10 & 11 & 12 \\
                   13 & 14 & 15 & 16
               \end{pmatrix}.

    Applying the dephasing channel to :math:`\rho` we have that

    .. math::
        \Phi(\rho) = \begin{pmatrix}
                         1 & 0 & 0 & 0 \\
                         0 & 6 & 0 & 0 \\
                         0 & 0 & 11 & 0 \\
                         0 & 0 & 0 & 16
                     \end{pmatrix}.

    This can be observed in :code:`toqito` as follows.

    >>> from toqito.channel_ops import apply_channel
    >>> from toqito.channels import dephasing
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    >>> )
    >>> apply_channel(test_input_mat, dephasing(4))
    [[ 1.,  0.,  0.,  0.],
     [ 0.,  6.,  0.,  0.],
     [ 0.,  0., 11.,  0.],
     [ 0.,  0.,  0., 16.]])

    We may also consider setting the parameter :code:`p = 0.5`.

    >>> from toqito.channel_ops import apply_channel
    >>> from toqito.channels import dephasing
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    >>> )
    >>> apply_channel(test_input_mat, dephasing(4, 0.5))
    [[17.5  0.   0.   0. ]
     [ 0.  20.   0.   0. ]
     [ 0.   0.  22.5  0. ]
     [ 0.   0.   0.  25. ]]

    References
    ==========
    .. [WatDeph18] Watrous, John.
        "The theory of quantum information."
        Section: "The completely dephasing channel".
        Cambridge University Press, 2018.

    :param dim: The dimensionality on which the channel acts.
    :param param_p: Default is 0.
    :return: The Choi matrix of the dephasing channel.
    """
    # Compute the Choi matrix of the dephasing channel.

    # Gives a sparse non-normalized state.
    psi = max_entangled(dim=dim, is_sparse=False, is_normalized=False)
    return (1 - param_p) * np.diag(np.diag(psi * psi.conj().T)) + param_p * (psi * psi.conj().T)



"""Schmidt rank of state."""
from typing import List, Union

import numpy as np


def schmidt_rank(vec: np.ndarray, dim: Union[int, List[int], np.ndarray] = None) -> float:
    r"""
    Compute the Schmidt rank [WikSR]_.

    For complex Euclidean spaces :math:`\mathcal{X}` and :math:`\mathcal{Y}`, a pure state
    :math:`u \in \mathcal{X} \otimes \mathcal{Y}` possesses an expansion of the form:

    .. math::
        u = \sum_{i} \lambda_i v_i w_i,

    where :math:`v_i \in \mathcal{X}` and :math:`w_i \in \mathcal{Y}` are orthonormal states.

    The Schmidt coefficients are calculated from

    .. math::
        A = \text{Tr}_{\mathcal{B}}(u^* u).

    The Schmidt rank is the number of non-zero eigenvalues of :math:`A`. The Schmidt rank allows us
    to determine if a given state is entangled or separable. For instance:

        - If the Schmidt rank is 1: The state is separable,
        - If the Schmidt rank > 1: The state is entangled.

    Compute the Schmidt rank of the vector :code:`vec`, assumed to live in bipartite space, where
    both subsystems have dimension equal to :code:`sqrt(len(vec))`.

    The dimension may be specified by the 1-by-2 vector :code:`dim` and the rank in that case is
    determined as the number of Schmidt coefficients larger than :code:`tol`.

    Examples
    ==========

    Computing the Schmidt rank of the entangled Bell state should yield a value greater than one.

    >>> from toqito.states import bell
    >>> from toqito.state_props import schmidt_rank
    >>> rho = bell(0).conj().T * bell(0)
    >>> schmidt_rank(rho)
    2

    Computing the Schmidt rank of the entangled singlet state should yield a value greater than
    :math:`1`.

    >>> from toqito.states import bell
    >>> from toqito.state_props import schmidt_rank
    >>> u = bell(2).conj().T * bell(2)
    >>> schmidt_rank(u)
    2

    Computing the Schmidt rank of a separable state should yield a value equal to :math:`1`.

    >>> from toqito.states import basis
    >>> from toqito.state_props import schmidt_rank
    >>> import numpy as np
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> e_00 = np.kron(e_0, e_0)
    >>> e_01 = np.kron(e_0, e_1)
    >>> e_10 = np.kron(e_1, e_0)
    >>> e_11 = np.kron(e_1, e_1)
    >>>
    >>> rho = 1 / 2 * (e_00 - e_01 - e_10 + e_11)
    >>> rho = rho.conj().T * rho
    >>> schmidt_rank(rho)
    1

    References
    ==========
    .. [WikSR] Wikipedia: Schmidt rank
        https://en.wikipedia.org/wiki/Schmidt_decomposition#Schmidt_rank_and_entanglement

    :param vec: A bipartite vector to have its Schmidt rank computed.
    :param dim: A 1-by-2 vector.
    :return: The Schmidt rank of vector :code:`vec`.
    """
    eps = np.finfo(float).eps
    slv = int(np.round(np.sqrt(len(vec))))
    if dim is None:
        dim = slv
    if isinstance(dim, int):
        dim = np.array([dim, len(vec) / dim], dtype=int)
        if np.abs(dim[1] - np.round(dim[1])) >= 2 * len(vec) * eps:
            raise ValueError(
                "Invalid: The value of `dim` must evenly divide "
                "`len(vec)`; please provide a `dim` array "
                "containing the dimensions of the subsystems"
            )
        dim[1] = np.round(dim[1])

    return np.linalg.matrix_rank(np.reshape(vec, dim[::-1]))



"""Hilbert-Schmidt metric."""
import numpy as np

from toqito.matrix_props import is_density


def hilbert_schmidt(rho: np.ndarray, sigma: np.ndarray) -> float:
    r"""
    Compute the Hilbert-Schmidt distance between two states [WikHS]_.

    The Hilbert-Schmidt distance between density operators :math:`\rho` and :math:`\sigma` is
    defined as

    .. math::
        D_{\text{HS}}(\rho, \sigma) = \text{Tr}((\rho - \sigma)^2) = \left\lVert \rho - \sigma
        \right\rVert_2^2.

    Examples
    ==========

    One may consider taking the Hilbert-Schmidt distance between two Bell states. In :code:`toqito`,
    one may accomplish this as

    >>> from toqito.states import bell
    >>> from toqito.state_metrics import hilbert_schmidt
    >>> rho = bell(0) * bell(0).conj().T
    >>> sigma = bell(3) * bell(3).conj().T
    >>> hilbert_schmidt(rho, sigma)
    1

    References
    ==========
    .. [WikHS] Wikipedia: Hilbert-Schmidt operator.
        https://en.wikipedia.org/wiki/Hilbert%E2%80%93Schmidt_operator

    :param rho: An input matrix.
    :param sigma: An input matrix.
    :return: The Hilbert-Schmidt distance between `rho` and `sigma`.
    """
    if not is_density(rho) or not is_density(sigma):
        raise ValueError("Hilbert-Schmidt is only defined for density " "operators.")
    return np.linalg.norm(rho - sigma, ord=2) ** 2



"""State distinguishability."""
from typing import List

import cvxpy
import numpy as np
from .state_helper import __is_states_valid, __is_probs_valid


def state_distinguishability(
    states: List[np.ndarray], probs: List[float] = None, dist_method: str = "min-error"
) -> float:
    r"""
    Compute probability of state distinguishability [ELD03]_.

    The "quantum state distinguishability" problem involves a collection of :math:`n` quantum states

    .. math::
        \rho = \{ \rho_0, \ldots, \rho_n \},

    as well as a list of corresponding probabilities

    .. math::
        p = \{ p_0, \ldots, p_n \}.

    Alice chooses :math:`i` with probability :math:`p_i` and creates the state :math:`\rho_i` Bob
    wants to guess which state he was given from the collection of states.

    One can specify the distinguishability method using the :code:`dist_method` argument.

    For :code:`dist_method = "min-error"`, this is the default method that yields the probability of
    distinguishing quantum states that minimize the probability of error.

    For :code:`dist_method = "unambiguous"`, Alice and Bob never provide an incorrect answer,
    although it is possible that their answer is inconclusive.

    When :code:`dist_method = "min-error"`, this function implements the following semidefinite
    program that provides the optimal probability with which Bob can conduct quantum state
    distinguishability.

    .. math::
        \begin{align*}
            \text{maximize:} \quad & \sum_{i=0}^n p_i \langle M_i, \rho_i \rangle \\
            \text{subject to:} \quad & M_0 + \ldots + M_n = \mathbb{I},\\
                                     & M_0, \ldots, M_n \geq 0
        \end{align*}

    When :code:`dist_method = "unambiguous"`, this function implements the following semidefinite
    program that provides the optimal probability with which Bob can conduct unambiguous quantum
    state distinguishability.

    .. math::
        \begin{align*}
            \text{maximize:} \quad & \sum_{i=0}^n p_i \langle M_i, \rho_i \rangle \\
            \text{subject to:} \quad & M_0 + \ldots + M_{n+1} = \mathbb{I},\\
                                     & \langle M_i, \rho_j \rangle = 0,
                                       \quad 1 \leq i, j \leq n, \quad i \not= j.
                                     & M_0, \ldots, M_n \geq 0
        \end{align*}

    Examples
    ==========

    State distinguishability for two state density matrices.

    >>> from toqito.states import basis, bell
    >>> from toqito.state_opt import state_distinguishability
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> e_00 = e_0 * e_0.conj().T
    >>> e_11 = e_1 * e_1.conj().T
    >>> states = [e_00, e_11]
    >>> probs = [1 / 2, 1 / 2]
    >>> res = state_distinguishability(states, probs)
    0.5000000000006083

    References
    ==========
    .. [ELD03] Eldar, Yonina C.
        "A semidefinite programming approach to optimal unambiguous
        discrimination of quantum states."
        IEEE Transactions on information theory 49.2 (2003): 446-456.
        https://arxiv.org/abs/quant-ph/0206093

    :param states: A list of states provided as either matrices or vectors.
    :param probs: Respective list of probabilities each state is selected.
    :param dist_method: Method of distinguishing to use.
    :return: The optimal probability with which Bob can distinguish the state.
    """
    obj_func = []
    measurements = []
    constraints = []

    __is_states_valid(states)
    if probs is None:
        probs = [1 / len(states)] * len(states)
    __is_probs_valid(probs)

    dim_x, dim_y = states[0].shape

    # The variable `states` is provided as a list of vectors. Transform them
    # into density matrices.
    if dim_y == 1:
        for i, state_ket in enumerate(states):
            states[i] = state_ket * state_ket.conj().T

    # Unambiguous state discrimination has an additional constraint on the states and measurements.
    if dist_method == "unambiguous":

        # Note we have one additional measurement operator in the unambiguous case.
        for i in range(len(states) + 1):
            measurements.append(cvxpy.Variable((dim_x, dim_x), PSD=True))

        # This is an extra condition required for the unambiguous case.
        for i, _ in enumerate(states):
            for j, _ in enumerate(states):
                if i != j:
                    constraints.append(cvxpy.trace(states[j].conj().T @ measurements[i]) == 0)

    if dist_method == "min-error":
        for i, _ in enumerate(states):
            measurements.append(cvxpy.Variable((dim_x, dim_x), PSD=True))

    # Objective function is the inner product between the states and measurements.
    for i, _ in enumerate(states):
        obj_func.append(probs[i] * cvxpy.trace(states[i].conj().T @ measurements[i]))

    constraints.append(sum(measurements) == np.identity(dim_x))

    objective = cvxpy.Maximize(sum(obj_func))
    problem = cvxpy.Problem(objective, constraints)
    sol_default = problem.solve()

    return sol_default



"""Tests for von_neumann_entropy."""
import numpy as np

from toqito.state_props import von_neumann_entropy
from toqito.states import bell
from toqito.states import max_mixed


def test_von_neumann_entropy_bell_state():
    """Entangled state von Neumann entropy should be zero."""
    rho = bell(0) * bell(0).conj().T
    res = von_neumann_entropy(rho)
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_von_neumann_entropy_max_mixed_statre():
    """Von Neumann entropy of the maximally mixed state should be one."""
    res = von_neumann_entropy(max_mixed(2, is_sparse=False))
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_von_neumann_non_density_matrix():
    r"""Test von Neumann entropy on non-density matrix."""
    rho = np.array([[1, 2], [3, 4]])

    with np.testing.assert_raises(ValueError):
        von_neumann_entropy(rho)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Fidelity metric."""
import cvxpy
import scipy
import numpy as np

from toqito.matrix_props import is_density


def fidelity(rho: np.ndarray, sigma: np.ndarray) -> float:
    r"""
    Compute the fidelity of two density matrices [WikFid]_.

    Calculate the fidelity between the two density matrices :code:`rho` and :code:`sigma`, defined
    by:

    .. math::
        ||\sqrt(\rho) * \sqrt(\sigma)||_1,

    where :math:`|| \cdot ||_1` denotes the trace norm. The return is a value between :math:`0` and
    :math:`1`, with :math:`0` corresponding to matrices :code:`rho` and :code:`sigma` with
    orthogonal support, and :math:`1` corresponding to the case :code:`rho = sigma`.

    Examples
    ==========

    Consider the following Bell state

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \in \mathcal{X}.

    The corresponding density matrix of :math:`u` may be calculated by:

    .. math::
        \rho = u u^* = \frac{1}{2} \begin{pmatrix}
                         1 & 0 & 0 & 1 \\
                         0 & 0 & 0 & 0 \\
                         0 & 0 & 0 & 0 \\
                         1 & 0 & 0 & 1
                       \end{pmatrix} \in \text{D}(\mathcal{X}).

    In the event where we calculate the fidelity between states that are identical, we should obtain
    the value of :math:`1`. This can be observed in :code:`toqito` as follows.

    >>> from toqito.state_metrics import fidelity
    >>> import numpy as np
    >>> rho = 1 / 2 * np.array(
    >>>     [[1, 0, 0, 1],
    >>>      [0, 0, 0, 0],
    >>>      [0, 0, 0, 0],
    >>>      [1, 0, 0, 1]]
    >>> )
    >>> sigma = rho
    >>> fidelity(rho, sigma)
    1.0000000000000002

    References
    ==========
    .. [WikFid] Wikipedia: Fidelity of quantum states
        https://en.wikipedia.org/wiki/Fidelity_of_quantum_states

    :param rho: Density operator.
    :param sigma: Density operator.
    :return: The fidelity between :code:`rho` and :code:`sigma`.
    """
    # Perform some error checking.
    if not np.all(rho.shape == sigma.shape):
        raise ValueError("InvalidDim: `rho` and `sigma` must be matrices of the same size.")

    # If `rho` or `sigma` is a cvxpy variable then compute fidelity via
    # semidefinite programming, so that this function can be used in the
    # objective function or constraints of other cvxpy optimization problems.
    if isinstance(rho, cvxpy.atoms.affine.vstack.Vstack) or isinstance(
        sigma, cvxpy.atoms.affine.vstack.Vstack
    ):
        z_var = cvxpy.Variable(rho.shape, complex=True)
        objective = cvxpy.Maximize(cvxpy.real(cvxpy.trace(z_var + z_var.H)))
        constraints = [cvxpy.bmat([[rho, z_var], [z_var.H, sigma]]) >> 0]
        problem = cvxpy.Problem(objective, constraints)

        return 1 / 2 * problem.solve()

    if not is_density(rho) or not is_density(sigma):
        raise ValueError("Fidelity is only defined for density operators.")

    # If `rho` or `sigma` are *not* cvxpy variables, compute fidelity normally,
    # since this is much faster.
    sq_rho = scipy.linalg.sqrtm(rho)
    sq_fid = scipy.linalg.sqrtm(sq_rho @ sigma @ sq_rho)
    return np.real(np.trace(sq_fid))



"""Test schmidt_decomposition."""
import numpy as np

from toqito.state_ops import schmidt_decomposition
from toqito.states import max_entangled


def test_schmidt_decomp_max_ent():
    """Schmidt decomposition of the 3-D maximally entangled state."""
    singular_vals, u_mat, vt_mat = schmidt_decomposition(max_entangled(3))

    expected_u_mat = np.identity(3)
    expected_vt_mat = np.identity(3)
    expected_singular_vals = 1 / np.sqrt(3) * np.array([[1], [1], [1]])

    bool_mat = np.isclose(expected_u_mat, u_mat)
    np.testing.assert_equal(np.all(bool_mat), True)

    bool_mat = np.isclose(expected_vt_mat, vt_mat)
    np.testing.assert_equal(np.all(bool_mat), True)

    bool_mat = np.isclose(expected_singular_vals, singular_vals)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_schmidt_decomp_dim_list():
    """Schmidt decomposition with list specifying dimension."""
    singular_vals, u_mat, vt_mat = schmidt_decomposition(max_entangled(3), dim=[3, 3])

    expected_u_mat = np.identity(3)
    expected_vt_mat = np.identity(3)
    expected_singular_vals = 1 / np.sqrt(3) * np.array([[1], [1], [1]])

    bool_mat = np.isclose(expected_u_mat, u_mat)
    np.testing.assert_equal(np.all(bool_mat), True)

    bool_mat = np.isclose(expected_vt_mat, vt_mat)
    np.testing.assert_equal(np.all(bool_mat), True)

    bool_mat = np.isclose(expected_singular_vals, singular_vals)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Breuer state."""
import numpy as np

from toqito.perms import symmetric_projection
from toqito.states import max_entangled


def breuer(dim: int, lam: float) -> np.ndarray:
    r"""
    Produce a Breuer state [HPBreuer]_.

    Gives a Breuer bound entangled state for two qudits of local dimension :code:`dim`, with the
    :code:`lam` parameter describing the weight of the singlet component as described in [HPBreuer].

    This function was adapted from the QETLAB package.

    Examples
    ==========

    We can generate a Breuer state of dimension :math:`4` with weight :math:`0.1`. For any weight
    above :math:`0`, the state will be bound entangled, that is, it will satisfy the PPT criterion,
    but it will be entangled.

    >>> from toqito.states import breuer
    >>> breuer(2, 0.1)
    [[ 0.3  0.  -0.   0. ]
     [ 0.   0.2  0.1  0. ]
     [-0.   0.1  0.2 -0. ]
     [ 0.   0.  -0.   0.3]]

    References
    ==========
    .. [HPBreuer] H-P. Breuer. Optimal entanglement criterion for mixed quantum states.
       E-print: arXiv:quant-ph/0605036, 2006.

    :param dim: Dimension of the Breuer state.
    :param lam: The weight of the singlet component.
    :return: Breuer state of dimension :code:`dim` with weight :code:`lam`.
    """
    if dim % 2 == 1 or dim <= 0:
        raise ValueError(f"The value {dim} must be an even positive integer.")

    v_mat = np.fliplr(np.diag((-1) ** np.mod(np.arange(1, dim + 1), 2)))
    max_entangled(dim)
    psi = np.dot(np.kron(np.identity(dim), v_mat), max_entangled(dim))

    return lam * (psi * psi.conj().T) + (1 - lam) * 2 * symmetric_projection(dim) / (
        dim * (dim + 1)
    )



"""Test in_separable_ball."""
import numpy as np

from toqito.random import random_unitary
from toqito.state_props import in_separable_ball


def test_in_separable_ball_matrix_true():
    """Test matrix in separable ball returns True."""
    u_mat = random_unitary(4)
    lam = np.array([1, 1, 1, 0]) / 3
    rho = u_mat @ np.diag(lam) @ u_mat.conj().T
    np.testing.assert_equal(in_separable_ball(rho), True)


def test_in_separable_ball_matrix_false():
    """Test matrix not in separable ball returns False."""
    u_mat = random_unitary(4)
    lam = np.array([1.01, 1, 0.99, 0]) / 3
    rho = u_mat @ np.diag(lam) @ u_mat.conj().T
    np.testing.assert_equal(in_separable_ball(rho), False)


def test_in_separable_ball_trace_lt_dim():
    """Test for case when trace of matrix is less than the largest dim."""
    rho = np.zeros((4, 4))
    np.testing.assert_equal(in_separable_ball(rho), False)


def test_in_separable_ball_eigs_false():
    """Test eigenvalues of matrix not in separable ball returns False."""
    u_mat = random_unitary(4)
    lam = np.array([1.01, 1, 0.99, 0]) / 3
    rho = u_mat @ np.diag(lam) @ u_mat.conj().T

    eig_vals = np.linalg.eigvalsh(rho)
    np.testing.assert_equal(in_separable_ball(eig_vals), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test symmetric_extension_hierarchy."""
import numpy as np
import pytest

from toqito.perms import swap
from toqito.state_opt import symmetric_extension_hierarchy
from toqito.states import basis, bell


def test_symmetric_extension_hierarchy_four_bell_density_matrices():
    """Symmetric extension hierarchy for four Bell density matrices."""
    states = [
        bell(0) * bell(0).conj().T,
        bell(1) * bell(1).conj().T,
        bell(2) * bell(2).conj().T,
        bell(3) * bell(3).conj().T,
    ]
    res = symmetric_extension_hierarchy(states=states, probs=None, level=2)
    np.testing.assert_equal(np.isclose(res, 1 / 2), True)


def test_symmetric_extension_hierarchy_four_bell_states():
    """Symmetric extension hierarchy for four Bell states."""
    states = [bell(0), bell(1), bell(2), bell(3)]
    res = symmetric_extension_hierarchy(states=states, probs=None, level=2)
    np.testing.assert_equal(np.isclose(res, 1 / 2), True)


def test_symmetric_extension_hierarchy_four_bell_with_resource_state_lvl_1():
    """Level 1 of hierarchy for four Bell states and resource state."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00, e_11 = np.kron(e_0, e_0), np.kron(e_1, e_1)

    eps = 1 / 2
    eps_state = np.sqrt((1 + eps) / 2) * e_00 + np.sqrt((1 - eps) / 2) * e_11
    eps_dm = eps_state * eps_state.conj().T

    states = [
        np.kron(bell(0) * bell(0).conj().T, eps_dm),
        np.kron(bell(1) * bell(1).conj().T, eps_dm),
        np.kron(bell(2) * bell(2).conj().T, eps_dm),
        np.kron(bell(3) * bell(3).conj().T, eps_dm),
    ]

    # Ensure we are checking the correct partition of the states.
    states = [
        swap(states[0], [2, 3], [2, 2, 2, 2]),
        swap(states[1], [2, 3], [2, 2, 2, 2]),
        swap(states[2], [2, 3], [2, 2, 2, 2]),
        swap(states[3], [2, 3], [2, 2, 2, 2]),
    ]

    # Level 1 of the hierarchy should be identical to the known PPT value
    # for this case.
    res = symmetric_extension_hierarchy(states=states, probs=None, level=1)
    exp_res = 1 / 2 * (1 + np.sqrt(1 - eps ** 2))

    np.testing.assert_equal(np.isclose(res, exp_res), True)


@pytest.mark.skip(reason="This test takes too much time.")  # pylint: disable=not-callable
def test_symmetric_extension_hierarchy_four_bell_with_resource_state():
    """Symmetric extension hierarchy for four Bell states and resource state."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00, e_11 = np.kron(e_0, e_0), np.kron(e_1, e_1)

    eps = 1 / 2
    eps_state = np.sqrt((1 + eps) / 2) * e_00 + np.sqrt((1 - eps) / 2) * e_11
    eps_dm = eps_state * eps_state.conj().T

    states = [
        np.kron(bell(0) * bell(0).conj().T, eps_dm),
        np.kron(bell(1) * bell(1).conj().T, eps_dm),
        np.kron(bell(2) * bell(2).conj().T, eps_dm),
        np.kron(bell(3) * bell(3).conj().T, eps_dm),
    ]

    # Ensure we are checking the correct partition of the states.
    states = [
        swap(states[0], [2, 3], [2, 2, 2, 2]),
        swap(states[1], [2, 3], [2, 2, 2, 2]),
        swap(states[2], [2, 3], [2, 2, 2, 2]),
        swap(states[3], [2, 3], [2, 2, 2, 2]),
    ]

    res = symmetric_extension_hierarchy(states=states, probs=None, level=2)
    exp_res = 1 / 2 * (1 + np.sqrt(1 - eps ** 2))

    np.testing.assert_equal(np.isclose(res, exp_res), True)


@pytest.mark.skip(reason="This test takes too much time.")  # pylint: disable=not-callable
def test_symmetric_extension_hierarchy_three_bell_with_resource_state():
    """Symmetric extension hierarchy for three Bell and resource state."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00, e_11 = np.kron(e_0, e_0), np.kron(e_1, e_1)

    eps = 1 / 2
    eps_state = np.sqrt((1 + eps) / 2) * e_00 + np.sqrt((1 - eps) / 2) * e_11
    eps_dm = eps_state * eps_state.conj().T

    states = [
        np.kron(bell(0) * bell(0).conj().T, eps_dm),
        np.kron(bell(1) * bell(1).conj().T, eps_dm),
        np.kron(bell(2) * bell(2).conj().T, eps_dm),
    ]

    # Ensure we are checking the correct partition of the states.
    states = [
        swap(states[0], [2, 3], [2, 2, 2, 2]),
        swap(states[1], [2, 3], [2, 2, 2, 2]),
        swap(states[2], [2, 3], [2, 2, 2, 2]),
    ]

    res = symmetric_extension_hierarchy(states=states, probs=None, level=2)
    np.testing.assert_equal(np.isclose(res, 0.9583057), True)


def test_invalid_symmetric_extension_hierarchy_probs():
    """Invalid probability vector for symmetric extension hierarchy."""
    with np.testing.assert_raises(ValueError):
        rho1 = bell(0) * bell(0).conj().T
        rho2 = bell(1) * bell(1).conj().T
        states = [rho1, rho2]
        symmetric_extension_hierarchy(states, [1, 2, 3])


def test_invalid_symmetric_extension_hierarchy_states():
    """Invalid number of states for symmetric_extension_hierarchy."""
    with np.testing.assert_raises(ValueError):
        states = []
        symmetric_extension_hierarchy(states)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test fourier."""
import numpy as np

from toqito.matrices import fourier


def test_fourier_dim_2():
    """Fourier matrix of dimension 2."""
    expected_res = np.array([[1 / np.sqrt(2), 1 / np.sqrt(2)], [1 / np.sqrt(2), -1 / np.sqrt(2)]])

    res = fourier(2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Shift matrix."""
import numpy as np


def shift(dim: int) -> np.ndarray:
    r"""
    Produce a `dim`-by-`dim` shift matrix [WikShift]_.

    Returns the shift matrix of dimension `dim` described in [WikShift]_. The
    shift matrix generates the following `dim`-by-`dim` matrix:

    .. math::
        \Sigma_{1, d} = \begin{pmatrix}
                        0 & 0 & 0 & \ldots & 0 & 1 \\
                        1 & 0 & 0 & \ldots & 0 & 0 \\
                        0 & 1 & 0 & \ldots & 0 & 0 \\
                        0 & 0 & 1 & \ldots & 0 & 0 \\
                        \vdots & \vdots & \vdots & \ddots & \vdots & \vdots \\
                        0 & 0 & 0 & \ldots & 1 & 0
                    \end{pmatrix}

    The shift matrix is primarily used in the construction of the generalized
    Pauli operators.

    Examples
    ==========

    The shift matrix generated from :math:`d = 3` yields the following matrix:

    .. math::
        \Sigma_{1, 3} =
        \begin{pmatrix}
            0 & 0 & 1 \\
            1 & 0 & 0 \\
            0 & 1 & 0
        \end{pmatrix}

    >>> from toqito.matrices import shift
    >>> shift(3)
    [[0., 0., 1.],
     [1., 0., 0.],
     [0., 1., 0.]]

    References
    ==========
    .. [WikShift] Wikipedia: Generalizations of Pauli matrices
        (https://en.wikipedia.org/wiki/Generalizations_of_Pauli_matrices.

    :param dim: Dimension of the matrix.
    :return: `dim`-by-`dim` shift matrix.
    """
    shift_mat = np.identity(dim)
    shift_mat = np.roll(shift_mat, -1)
    shift_mat[:, -1] = np.array([0] * dim)
    shift_mat[0, -1] = 1

    return shift_mat



"""Test unique_perms."""
import numpy as np

from toqito.perms import unique_perms


def test_unique_perms_len():
    """Checks the number of unique perms."""
    vec = [1, 1, 2, 2, 1, 2, 1, 3, 3, 3]
    np.testing.assert_equal(len(list(unique_perms(vec))), 4200)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Determine whether there exists a symmetric extension for a given quantum state."""
import numpy as np

from toqito.channels import partial_trace
from toqito.matrix_props import is_positive_semidefinite
from toqito.state_opt import symmetric_extension_hierarchy
from toqito.state_props import is_ppt


def has_symmetric_extension(rho, level=2, dim=None, ppt=True, tol=1e-4) -> bool:
    r"""
    Determine whether there exists a symmetric extension for a given quantum state. [DPS02]_.

    Determining whether an operator possesses a symmetric extension at some level :code:`level`
    can be used as a check to determine if the operator is entangled or not.

    This function was adapted from QETLAB.

    Examples
    ==========

    2-qubit symmetric extension:

    In [CJKLZB14]_, it was shown that a 2-qubit state :math:`\rho_{AB}` has a
    symmetric extension if and only if

    .. math::
        \text{Tr}(\rho_B^2) \geq \text{Tr}(\rho_{AB}^2) - 4 \sqrt{\text{det}(\rho_{AB})}.

    This closed-form equation is much quicker to check than running the semidefinite program.

    >>> import numpy as np
    >>> from toqito.state_props import has_symmetric_extension
    >>> rho = np.array([[1, 0, 0, -1],
    >>>                 [0, 1, 1/2, 0],
    >>>                 [0, 1/2, 1, 0],
    >>>                 [-1, 0, 0, 1]])
    >>> # Show the closed-form equation holds
    >>> np.trace(partial_trace(rho, 1)**2) >= np.trace(rho**2) - 4 * np.sqrt(np.linalg.det(rho))
    True
    >>> # Now show that the `has_symmetric_extension` function recognizes this case.
    >>> has_symmetric_extension(rho)
    True

    Higher qubit systems:

    Consider a density operator corresponding to one of the Bell states.

    .. math::
        \rho = \frac{1}{2} \begin{pmatrix}
                            1 & 0 & 0 & 1 \\
                            0 & 0 & 0 & 0 \\
                            0 & 0 & 0 & 0 \\
                            1 & 0 & 0 & 1
                           \end{pmatrix}

    To make this state over more than just two qubits, let's construct the following state

    .. math::
        \sigma = \rho \otimes \rho.

    As the state :math:`\sigma` is entangled, there should not exist a symmetric extension at some
    level. We see this being the case for a relatively low level of the hierachy.

    >>> import numpy as np
    >>> from toqito.states import bell
    >>> from toqito.state_props import has_symmetric_extension
    >>>
    >>> rho = bell(0) * bell(0).conj().T
    >>> sigma = np.kron(rho, rho)
    >>> has_symmetric_extension(sigma)
    False

    References
    ==========
    .. [DPS02] Doherty, Andrew C., Pablo A. Parrilo, and Federico M. Spedalieri.
        "Distinguishing separable and entangled states."
        Physical Review Letters 88.18 (2002): 187904.
        https://arxiv.org/abs/quant-ph/0112007

    .. [CJKLZB14] Chen, J., Ji, Z., Kribs, D., Lütkenhaus, N., & Zeng, B.
        "Symmetric extension of two-qubit states."
        Physical Review A 90.3 (2014): 032318.
        https://arxiv.org/abs/1310.3530

    :param rho: A matrix or vector.
    :param level: Level of the hierarchy to compute.
    :param dim: The default has both subsystems of equal dimension.
    :param ppt: If :code:`True`, this enforces that the symmetric extension must be PPT.
    :param tol: Tolerance when determining whether a symmetric extension exists.
    :return: :code:`True` if :code:`mat` has a symmetric extension; :code:`False` otherwise.
    """
    len_mat = rho.shape[1]

    # Set default dimension if none was provided.
    if dim is None:
        dim = int(np.round(np.sqrt(len_mat)))

    # Allow the user to enter in a single integer for dimension.
    if isinstance(dim, int):
        dim = np.array([dim, len_mat / dim])
        if np.abs(dim[1] - np.round(dim[1])) >= 2 * len_mat * np.finfo(float).eps:
            raise ValueError(
                "If `dim` is a scalar, it must evenly divide the length of the matrix."
            )
        dim[1] = int(np.round(dim[1]))

    dim_x, dim_y = int(dim[0]), int(dim[1])
    # In certain situations, we don't need semidefinite programming.
    if level == 1 or len_mat <= 6 and ppt:
        if not ppt:
            # In some cases, the problem is *really* trivial.
            return is_positive_semidefinite(rho)

        # In this case, all they asked for is a 1-copy PPT symmetric extension
        # (i.e., they're asking if the state is PPT).
        return is_ppt(rho, 2, dim) and is_positive_semidefinite(rho)

    # In the 2-qubit case, an analytic formula is known for whether or not a state has a
    # (2-copy, non-PPT) symmetric extension that is much faster to use than semidefinite
    # programming [CJKLZB14]_.
    if level == 2 and not ppt and dim_x == 2 and dim_y == 2:
        return np.trace(np.linalg.matrix_power(partial_trace(rho, 1), 2)) >= np.trace(
            np.linalg.matrix_power(rho, 2)
        ) - 4 * np.sqrt(np.linalg.det(rho))

    # Otherwise, use semidefinite programming to find a symmetric extension.
    # If the optimal value of the symmetric extension hierarchy is equal to 1,
    # this indicates that there does not exist a symmetric extension at
    # level :code:`level`.
    return not np.isclose(
        (1 - min(symmetric_extension_hierarchy([rho], probs=None, level=level), 1)),
        0,
        atol=tol,
    )



"""Test random_state_vector."""
import numpy as np

from toqito.random import random_state_vector
from toqito.state_props import is_pure, purity


def test_random_state_vector_complex_state_purity():
    """Check that complex state vector from random state vector is pure."""
    vec = random_state_vector(2)
    mat = vec * vec.conj().T
    np.testing.assert_equal(is_pure(mat), True)


def test_random_state_vector_complex_state_purity_k_param():
    """Check that complex state vector with k_param > 0."""
    vec = random_state_vector(2, False, 1).reshape(-1, 1)
    mat = vec * vec.conj().T
    np.testing.assert_equal(np.isclose(purity(mat), 1), True)


def test_random_state_vector_complex_state_purity_k_param_dim_list():
    """Check that complex state vector with k_param > 0 and dim list."""
    vec = random_state_vector([2, 2], False, 1).reshape(-1, 1)
    mat = vec * vec.conj().T
    np.testing.assert_equal(np.isclose(purity(mat), 1), True)


def test_random_state_vector_real_state_purity_with_k_param():
    """Check that real state vector with k_param > 0."""
    vec = random_state_vector(2, True, 1).reshape(-1, 1)
    mat = vec * vec.conj().T
    np.testing.assert_equal(np.isclose(purity(mat), 1), True)


def test_random_state_vector_real_state_purity():
    """Check that real state vector from random state vector is pure."""
    vec = random_state_vector(2, True).reshape(-1, 1)
    mat = vec * vec.conj().T
    np.testing.assert_equal(np.isclose(purity(mat), 1), True)


if __name__ == "__main__":
    np.testing.run_module_suite()
